"""
The OpenMP package does contain all code templates required for the openMP
code generation in ANNarchy.

BaseTemplates:

    defines the basic defintions common to all sparse matrix formates, e. g. projection header

[FORMAT]_SingleThread:

    defines the format specific defintions for the currently available formats:

        * LIL: list-in-list
        * BSR: blocked sparse row
        * COO: coordinate
        * CSR: compressed sparse row
        * ELL: ELLPACK/ITPACK
        * ELLR: ELLPACK format with row-length array
        * SELL: sliced ELLPACK format
        * Dense: a full matrix representation

    there are some special purpose implementations:

        * CSR_T: compressed sparse row (transposed)
        * Dense_T: dense (transposed)
        * LIL_P: a partitioned LIL representation
"""
from . import LIL as LIL_OpenMP
from . import LIL_P as LIL_Sliced_OpenMP
from . import COO as COO_OpenMP
from . import DIA as DIA_OpenMP
from . import BSR as BSR_OpenMP
from . import CSR as CSR_OpenMP
from . import CSR_P as CSR_Sliced_OpenMP
from . import CSR_T as CSR_T_OpenMP
from . import CSR_T_P as CSR_T_Sliced_OpenMP
from . import ELL as ELL_OpenMP
from . import ELLR as ELLR_OpenMP
from . import SELL as SELL_OpenMP
from . import Dense as Dense_OpenMP
from . import Dense_T as Dense_T_OpenMP

__all__ = [
    "BaseTemplates",
    "LIL_OpenMP", "LIL_Sliced_OpenMP",
    "BSR_OpenMP",
    "COO_OpenMP",
    "DIA_OpenMP",
    "CSR_OpenMP", "CSR_Sliced_OpenMP", "CSR_T_OpenMP", "CSR_T_Sliced_OpenMP",
    "ELL_OpenMP",
    "ELLR_OpenMP",
    "SELL_OpenMP",
    "Dense_OpenMP", "Dense_T_OpenMP"
]



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
# (directly imported by CodeGenerator if needed)
additional_global_functions = """
// Alternative implementation for Keplar and upwards
// https://developer.nvidia.com/blog/faster-parallel-reductions-kepler/
#define FULL_WARP_MASK 0xFFFFFFFF
template<class ValueType, unsigned int WARP_SIZE>
__device__ ValueType warp_reduce (ValueType val)
{
    for(int offset = WARP_SIZE/2; offset > 0; offset /= 2)
        val += __shfl_down_sync(FULL_WARP_MASK, val, offset, 32);

    return val;
}
"""

launch_config = {
    'init': """
        _threads_per_block = 64;
        _nb_blocks = std::min<unsigned int>(ceil(static_cast<double>(nb_dendrites())/static_cast<double>(_threads_per_block)), 65535);
    
    #ifdef _DEBUG
        std::cout << "Kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
""",
    'update': """
    #ifdef _DEBUG
        std::cout << "The CSR vector implementation is fixed to a specific thread configuration." << std::endl;
    #endif
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'semiglobal': """
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'global': {
        'parameter': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
""",
        'variable': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
"""
    }
}

attribute_cpp_init = {
    'local': """
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_vector_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'global': {
        'parameter': """
        // Global %(attr_type)s %(name)s
        %(name)s = 0.0;
    """,
        'variable': """
        // Global %(attr_type)s %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
"""
    }
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'semiglobal': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), num_non_zeros_ * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), post_ranks_.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, &%(name)s, sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
"""
}

attribute_device_to_host = {
    'local': """
        // %(name)s: local
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, num_non_zeros_ * sizeof( %(type)s ), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'semiglobal': """
            // %(name)s: semiglobal
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, post_ranks_.size() * sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
""",
    'global': """
            // %(name)s: global
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( &%(name)s, gpu_%(name)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
"""
}

delay = {
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;""",
        'pyx_struct': """
        # Non-uniform delay
        int delay""",
        'init': "delay = delays[0][0];",
        'pyx_wrapper_init': """
        proj%(id_proj)s.delay = syn.uniform_delay""",
        'pyx_wrapper_accessor': """
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay
    def set_delay(self, value):
        print("set delay", value)
        proj%(id_proj)s.delay = value
"""
    },
    'nonuniform_rate_coded': {
        'declare': """
    // Non-uniform delay
    std::vector< std::vector< int > > delay ;""",
        'pyx_struct': """
        # Non-uniform delay
        vector[vector[int]] delay""",
    
        'init': "",

        'pyx_wrapper_init': """
        proj%(id_proj)s.delay = syn.delay""",
    
        'pyx_wrapper_accessor': """
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay[idx]
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""
    },
    'nonuniform_spiking': None
}

event_driven = {
    'declare': """
    std::vector< long > _last_event;
    long* _gpu_last_event;
""",
    'cpp_init': """
    _last_event = init_matrix_variable<long>(-10000);
    _gpu_last_event = init_matrix_variable_gpu<long>(_last_event);    
""",
    'pyx_struct': """
        vector[long] _last_event
""",
    'pyx_wrapper_init':
"""
        proj%(id_proj)s._last_event = vector[long]( syn._matrix.num_elements(), -10000)
"""
}

rate_psp_kernel = {
    # Adapted from Bell and Garland 2008, taken from https://code.google.com/archive/p/cusp-library/downloads (26. mar. 2018)
    #
    #   HD (2019): I needed to add a volatile variable, otherwise the results were wrong ...
    #   HD (2020): I replaced the local reduction by a warp primitive introduced with CUDA 9.0
    'device_kernel': {
        'sum':"""
template<unsigned int BLOCK_SIZE, unsigned int WARP_SIZE>
__global__ void cu_proj%(id_proj)s_psp(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s ) {
    __shared__ %(size_type)s ptrs[BLOCK_SIZE/WARP_SIZE][2];

    const %(idx_type)s thread_id   = BLOCK_SIZE * blockIdx.x + threadIdx.x;  // global thread index
    const %(idx_type)s thread_lane = threadIdx.x & (WARP_SIZE-1);            // thread index within the warp
    const %(idx_type)s warp_id     = thread_id   / WARP_SIZE;                // global warp index
    const %(idx_type)s warp_lane   = threadIdx.x / WARP_SIZE;                // warp index within the CTA
    const %(idx_type)s num_warps   = (BLOCK_SIZE / WARP_SIZE) * gridDim.x;   // total number of active warps

    for(%(idx_type)s row = warp_id; row < post_size; row += num_warps){

        // use two threads to fetch Ap[row] and Ap[row+1]
        // this is considerably faster than the straightforward version
        if(thread_lane < 2)
            ptrs[warp_lane][thread_lane] = row_ptr[row + thread_lane];
        const %(size_type)s row_start = ptrs[warp_lane][0];                   //same as: row_start = Ap[row];
        const %(size_type)s row_end   = ptrs[warp_lane][1];                   //same as: row_end   = Ap[row+1];

        // compute local sum
        %(float_prec)s sum = 0;
        for(%(size_type)s j = row_start + thread_lane; j < row_end; j += WARP_SIZE)
            sum += %(psp)s

        // reduce local sums to row sum (ASSUME: warpsize 32)
        sum = warp_reduce<%(float_prec)s, WARP_SIZE>(sum);

        // first thread writes warp result
        if (thread_lane == 0)
            %(target_arg)s[rank_post[row]] = sum;
    }
}
"""
    },
    'invoke_kernel': """
void launch_proj%(id_proj)s_psp(const unsigned int nb_blocks, const unsigned int threads_per_block, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s ) {

    switch (threads_per_block)
    {
        case 64:
        {
            cu_proj%(id_proj)s_psp<64,32><<< nb_blocks, threads_per_block>>>(
                /* ranks and offsets */
                %(conn_args_call)s
                /* computation data */
                %(add_args_call)s
                /* result */
                %(target_arg_call)s
            );
        }break;

        default:
        {
            std::cerr << "The kernel configuration tpb = " << threads_per_block << " is not supported" << std::endl;
        }
    }
}
""",    
    'kernel_decl': """void launch_proj%(id_proj)s_psp(const unsigned int nb_blocks, const unsigned int threads_per_block, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s );
""",
    'host_call': """
    // proj%(id_proj)s: pop%(id_pre)s -> pop%(id_post)s
    if ( pop%(id_post)s->_active && proj%(id_proj)s->_transmission ) {
        const unsigned int BLOCK_SIZE = proj%(id_proj)s->_threads_per_block;
        const unsigned int WARPS_PER_BLOCK = BLOCK_SIZE / 32;
        const unsigned int MAX_BLOCKS = MAX_THREADS / BLOCK_SIZE;
        const unsigned int NUM_BLOCKS = std::min(MAX_BLOCKS, DIVIDE_INTO( proj%(id_proj)s->nb_dendrites(), WARPS_PER_BLOCK));

        launch_proj%(id_proj)s_psp(
            NUM_BLOCKS, BLOCK_SIZE,
            /* ranks and offsets */
            %(conn_args)s
            /* computation data */
            %(add_args)s
            /* result */
            %(target_arg)s
        );

    #ifdef _DEBUG
        auto err = cudaGetLastError();
        if ( err != cudaSuccess ) {
            std::cout << "cu_proj%(id_proj)s_psp: " << cudaGetErrorString(err) << std::endl;
        }
    #endif
    }
""",
    'thread_init': {
        'float': {
            'sum': "0.0f",
            'min': "FLT_MAX",
            'max': "FLT_MIN",
            'mean': "0.0f"
        },
        'double': {
            'sum': "0.0",
            'min': "DBL_MAX",
            'max': "DBL_MIN",
            'mean': "0.0"
        }
    }
}

conn_templates = {
    # connectivity representation
    'conn_header' : "const %(idx_type)s post_size, const %(idx_type)s* __restrict__  rank_post, const %(size_type)s* __restrict__ row_ptr, const %(idx_type)s* __restrict__ rank_pre",
    'conn_call' : "proj%(id_proj)s->nb_dendrites(), proj%(id_proj)s->gpu_post_rank, proj%(id_proj)s->gpu_row_ptr, proj%(id_proj)s->gpu_pre_rank",
    'conn_kernel': "post_size, rank_post, row_ptr, rank_pre",

    # launch config
    'launch_config': launch_config,

    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete':attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,
    'delay': delay,
    'event_driven': event_driven,

    # operations
    'rate_psp': rate_psp_kernel,
    'spike_transmission': {
        'event_driven': None,
        'continuous': None,
    },
    'synapse_update': {
        'global': None,
        'semiglobal': None,
        'local': None,
        'call': None
    },
    'post_event': None
}

conn_ids = {
    'local_index': "[j]",
    'semiglobal_index': '[i]',
    'global_index': '[0]',
    'pre_index': '[rank_pre[j]]',
    'post_index': '[rank_post[i]]',
    'pre_prefix': 'pre_',
    'post_prefix': 'post_',
    'delay_nu' : '[delay[j]-1]', # non-uniform delay
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from copy import deepcopy

from ANNarchy.core import Global
from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern import Messages
from ANNarchy.core.Projection import Projection
from ANNarchy.extensions.convolution import Convolution, Pooling

from .CopyTemplate import copy_proj_template, copy_sum_template
from .Utils import SharedSynapse

class Copy(SpecificProjection):
    """
    Creates a virtual projection reusing the weights and delays of an already-defined projection.

    Although the original projection can be learnable, this one can not. Changes in the original weights will be reflected in this projection. The only possible modifications are ``psp`` and ``operation``.

    The pre- and post-synaptic populations of both projections must have the same geometry.

    Example:

    ```python
    import ANNarchy as ann
    from ANNarchy.extensions.convolution import Copy

    net = ann.Network()

    pop1 = net.create(1000, ann.Izhikevich)
    pop2 = net.create(1000, ann.Izhikevich)
    pop3 = net.create(1000, ann.Izhikevich)

    proj = ann.Projection(pop1, pop2, "exc")
    proj.fixed_probability(0.1, 0.5)

    copy_proj = Copy(pop1, pop3, "exc")
    copy_proj.copy(proj)
    ```

    :param pre: pre-synaptic population (either its name or a ``Population`` object).
    :param post: post-synaptic population (either its name or a ``Population`` object).
    :param target: type of the connection
    :param psp: continuous influence of a single synapse on the post-synaptic neuron (default for rate-coded: ``w*pre.r``).
    :param operation: operation (sum, max, min, mean) performed by the kernel (default: sum).

    """
    def __init__(self, pre, post, target, psp="pre.r * w", operation="sum", name=None, copied=False, net_id=0):

        # Create the description, but it will not be used for generation
        SpecificProjection.__init__(
            self,
            pre=pre,
            post=post,
            target=target,
            synapse = SharedSynapse(psp=psp, operation=operation),
            name=name,
            copied=copied,
            net_id=net_id
        )

        self.connector_name = "Copy"
        self.connector_description = "Copy projection"

    def connect_copy(self, projection):
        return self.copy(projection)
    
    def copy(self, projection):
        """
        Instantiates the projection.

        :param projection: Existing projection to copy.
        """
        self.projection = projection

        # Sanity checks
        if not isinstance(self.projection, Projection):
            Messages._error('Copy: You must provide an existing projection to copy().')

        if isinstance(self.projection, (Convolution, Pooling)):
            Messages._error('Copy: You can only copy regular projections, not shared projections.')

        if not self.pre.geometry == self.projection.pre.geometry or not self.post.geometry == self.projection.post.geometry:
            Messages._error('Copy: When copying a projection, the geometries must be the same.')

        # Finish building the synapses
        self._create()

        return self

    def _copy(self, pre, post, net_id=None):
        "Returns a copy of the projection when creating networks. Internal use only."
        return Copy(
            pre=self.pre, post=self.post, target=self.target,
            psp=self.synapse_type.psp, operation=self.synapse_type.operation, 
            name=self.name, copied=True,
            net_id = self.net_id if not net_id else net_id
        )

    def _create(self):
        # create fake LIL object, just for compilation.
        try:
            from ANNarchy.cython_ext.Connector import LILConnectivity
        except Exception as e:
            Messages._print(e)
            Messages._error('ANNarchy was not successfully installed.')

        lil = LILConnectivity(dt=ConfigManager().get('dt', self.net_id))

        self.connector_name = "Copy"
        self.connector_description = "Copy projection"
        self._store_connectivity(self._load_from_lil, (lil, ), 0)

    def _connect(self, module):
        """
        Builds up dendrites either from list or dictionary. Called by instantiate().
        """
        if not self._connection_method:
            Messages._error('Copy: The projection between ' + self.pre.name + ' and ' + self.post.name + ' is declared but not connected.')

        # Create the Cython instance
        self.cyInstance = getattr(module, 'proj'+str(self.id)+'_wrapper')()

        return True

    def _generate(self):
        """
        Overrides default code generation. This function is called during the code generation procedure.
        """
        if _check_paradigm("openmp", self.net_id):
            self._generate_omp()
        elif _check_paradigm("cuda", self.net_id):
            self._generate_cuda()
        else:
            raise NotImplementedError

    def _generate_omp(self):
        """
        Code generation of CopyProjection object for the openMP paradigm.
        """
        # Set the projection specific parameters
        copy_proj_dict = deepcopy(copy_proj_template)
        for key, value in copy_proj_dict.items():
            value = value % {
                'id_proj': self.id,
                'id_copy': self.projection.id,
                'float_prec': ConfigManager().get('precision', self.net_id)
            }
            copy_proj_dict[key] = value

        # Update specific template
        self._specific_template.update(copy_proj_dict)

        # No attributes
        self._specific_template['declare_parameters_variables'] = ""
        self._specific_template['export_parameters_variables'] = ""
        self._specific_template['access_parameters_variables'] = ""

        # Override the monitor to avoid recording the weights
        self._specific_template['monitor_class'] = ""
        self._specific_template['monitor_export'] = ""
        self._specific_template['monitor_wrapper'] = ""

        self._specific_template['size_in_bytes'] = ""
        self._specific_template['clear_container'] = ""


        # OMP code if more then one thread
        if ConfigManager().get('num_threads', self.net_id) > 1:
            omp_code = '#pragma omp for private(sum)' if self.post.size > Global.OMP_MIN_NB_NEURONS else ''
        else:
            omp_code = ""

        # PSP
        psp = self.synapse_type.description['psp']['cpp']  % {
            'id_pre': self.pre.id,
            'id_post': self.post.id,
            'local_index':'[i][j]',
            'global_index': '[i]',
            'pre_index': f'[proj{self.projection.id}->pre_rank[i][j]]',
            'post_index': f'[proj{self.projection.id}->post_rank[i]]',
            'pre_prefix': 'pop'+str(self.pre.id)+'->',
            'post_prefix': 'pop'+str(self.post.id)+'->'}
        
        # Replace the weights with the target weights
        psp = psp.replace('rk_pre', f'proj{self.projection.id}->pre_rank[i][j]').replace(';', '')
        psp = psp.replace('w[i][j]', f'proj{self.projection.id}->w[i][j]')

        # Select template for operation to be performed: sum, max, min, mean
        try:
            sum_code = copy_sum_template[self.synapse_type.operation]
        except KeyError:
            Messages._error("CopyProjection: the operation ", self.synapse_type.operation, ' is not available.')

        # Finalize code
        self._specific_template['psp_code'] = sum_code % {
            'id_proj': self.id, 
            'target': self.target,
            'id_pre': self.pre.id, 'name_pre': self.pre.name,
            'id_post': self.post.id, 'name_post': self.post.name,
            'id': self.projection.id,
            'float_prec': ConfigManager().get('precision', self.net_id),
            'omp_code': omp_code,
            'psp': psp
        }

    def _generate_cuda(self):
        """
        Code generation of CopyProjection object for the CUDA paradigm.

        Note: currently not implemented (TODO HD)
        """
        raise NotImplementedError

    ##############################
    ## Override useless methods
    ##############################
    def _data(self):
        "Disable saving."
        desc = {}
        desc['attributes'] = self.attributes
        desc['parameters'] = self.parameters
        desc['variables'] = self.variables

        desc['dendrites'] = []
        desc['number_of_synapses'] = 0
        return desc

    def save_connectivity(self, filename):
        "Not available."
        Messages._warning('Copied projections can not be saved.')
    def save(self, filename):
        "Not available."
        Messages._warning('Copied projections can not be saved.')
    def load(self, filename):
        "Not available."
        Messages._warning('Copied projections can not be loaded.')
    def receptive_fields(self, variable = 'w', in_post_geometry = True):
        "Not available."
        Messages._warning('Copied projections can not display receptive fields.')
    def connectivity_matrix(self, fill=0.0):
        "Not available."
        Messages._warning('Copied projections can not display connectivity matrices.')



# distutils: language = c++
from libcpp.vector cimport vector
from libcpp cimport bool
from sympy.mpmath.matrices.matrices import _matrix

cdef class LILConnectivity:
    """
    Container for the ranks, weights and delays of a projection.
    """
    # Data
    cdef public vector[int] post_rank
    cdef public vector[vector[int]] pre_rank
    cdef public vector[vector[double]] w
    cdef public vector[vector[int]] delay

    # Attributes
    cdef public int max_delay
    cdef public int uniform_delay
    cdef public int size
    cdef public int nb_synapses
    cdef public double dt
    cdef public bool requires_sorting
    cdef public int last_added_idx

    # Insert methods
    cpdef add(self, int rk, r, w, d)
    cpdef push_back(self, int rk, vector[int] r, vector[double] w, vector[double] d)

    # Access methods
    cpdef int get_max_delay(self)
    cpdef int get_uniform_delay(self)

    # Matrix characteristics (auto-tuning)
    cpdef compute_average_row_length(self)
    cpdef compute_average_col_idx_gap(self)

    # Method to validate a LIL object
    cpdef validate(self)

    # pre-defined pattern
    cpdef all_to_all(self, pre, post, weights, delays, allow_self_connections)
    cpdef one_to_one(self, pre, post, weights, delays)
    cpdef fixed_probability(self, pre, post, probability, weights, delays, allow_self_connections)
    cpdef fixed_number_pre(self, pre, post, int number, weights, delays, allow_self_connections)
    cpdef fixed_number_post(self, pre, post, int number, weights, delays, allow_self_connections)
    cpdef gaussian(self, pre_pop, post_pop, float amp, float sigma, delays, limit, allow_self_connections)
    cpdef dog(self, pre_pop, post_pop, float amp_pos, float sigma_pos, float amp_neg, float sigma_neg, delays, limit, allow_self_connections)



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

global_operation_templates_st = {
    'max' : """
// Computes the maximum value of an array
%(type)s max_value(const %(type)s* array, int n)
{
    %(type)s max = array[0];
    for(int i=1; i<n; i++)
    {
        if(array[i] > max)
            max = array[i];
    }
    return max;
}
""",
    'min' : """
// Computes the minimum value of an array
%(type)s min_value(const %(type)s* array, int n)
{
    %(type)s min = array[0];
    for(int i=1; i<n; i++)
    {
        if(array[i] < min)
            min = array[i];
    }

    return min;
}
""",
    'mean' : """
// Computes the mean value of an array
%(type)s mean_value(const %(type)s* array, int n)
{
    %(type)s sum = array[0];
    for(int i=1; i<n; i++)
    {
        sum += array[i];
    }
    return sum/(%(type)s)n;
}
""",
    'norm1' : """
// Computes the L1-norm of an array
%(type)s norm1_value(const %(type)s* array, int n)
{
    %(type)s sum = fabs(array[0]);
    for(int i=1; i<n; i++)
    {
        sum += fabs(array[i]);
    }
    return sum;
}
""",
    'norm2' : """
// Computes the L2-norm (Euclidian) of an array
%(type)s norm2_value(const %(type)s* array, int n)
{
    %(type)s sum = array[0] * array[0];
    for(int i=1; i<n; i++)
    {
        sum += array[i] * array[i];
    }
    return sqrt(sum);
}
"""
}

global_operation_templates_st_extern = {
    'max': "extern %(type)s max_value(const %(type)s*, int);\n",
    'min': "extern %(type)s min_value(const %(type)s*, int);\n",
    'mean': "extern %(type)s mean_value(const %(type)s*, int);\n",
    'norm1': "extern %(type)s norm1_value(const %(type)s*, int);\n",
    'norm2': "extern %(type)s norm2_value(const %(type)s*, int);\n"
}

global_operation_templates_openmp = {
    'max' : """// Computes the maximum value of an array
%(type)s max_value(const %(type)s* array, const int beg, const int end)
{
    %(type)s max = array[beg];
    for(int i=beg+1; i<end; i++)
    {
        if(array[i] > max)
            max = array[i];
    }
    return max;
}
""",
    'min' : """
// Computes the minimum value of an array
%(type)s  min_value(const %(type)s* array, const int beg, const int end)
{
    %(type)s min = array[beg];
    for(int i=beg+1; i<end; i++)
    {
        if(array[i] < min)
            min = array[i];
    }
    return min;
}
""",
    'mean' : """
// Computes the mean value of an array
%(type)s mean_value(const %(type)s* array, const int beg, const int end, const int n)
{
    %(type)s sum = array[beg];
    for(int i=beg+1; i<end; i++)
    {
        sum += array[i];
    }
    return sum/static_cast<%(type)s>(n);
}

%(type)s mean_value(const %(type)s* array, const int beg, const int end) 
{
    return mean_value(array, beg, end, end);
}
""",
    'norm1' : """
// Computes the L1-norm of an array
%(type)s norm1_value(const %(type)s* array, const int beg, const int end)
{
    %(type)s sum = fabs(array[beg]);
    for(int i=beg+1; i<end; i++)
    {
        sum += fabs(array[i]);
    }
    return sum;
}
""",
    'norm2' : """
// Computes the L2-norm (Euclidian) of an array
%(type)s norm2_value(const %(type)s* array, const int beg, const int end)
{
    %(type)s sum = array[beg] * array[beg];
    for(int i=beg+1; i<end; i++)
    {
        sum += array[i] * array[i];
    }

    return sum;
}
"""
}

# Will be called if number of elements is too small to be computed in parallel
global_operation_templates_st_call = {
    'max': """#pragma omp task
{
    _max_%(var)s = max_value(%(var)s.data(), 0, %(var)s.size());
}
""",
    'min': """#pragma omp task
{
    _min_%(var)s = min_value(%(var)s.data(), 0, %(var)s.size());
}
""",
    'mean': """#pragma omp task
{
    _mean_%(var)s = mean_value(%(var)s.data(), 0, %(var)s.size());
}
""",
    'norm1': """#pragma omp task
{
    _norm1_%(var)s = norm1_value(%(var)s.data(), 0, %(var)s.size());
}
""",
    'norm2': """#pragma omp task
{
    _norm2_%(var)s = sqrt(norm2_value(%(var)s.data(), 0, %(var)s.size()));
}
"""
}

global_operation_templates_omp_call = {
    'max': """#pragma omp master
{
    _max_%(var)s = %(var)s[0];
}
auto local_max_%(var)s = max_value(%(var)s.data(), chunks_[tid], chunks_[tid+1]);
""",
    'min': """#pragma omp master
{
    _min_%(var)s = %(var)s[0];
}
auto local_min_%(var)s = min_value(%(var)s.data(), chunks_[tid], chunks_[tid+1]);
""",
    'mean': """#pragma omp master
{
    _mean_%(var)s = 0.0;
}
auto local_mean_%(var)s = mean_value(%(var)s.data(), chunks_[tid], chunks_[tid+1], %(var)s.size());
""",
    'norm1': """#pragma omp master
{
    _norm1_%(var)s = 0.0;
}
auto local_norm1_%(var)s = norm1_value(%(var)s.data(), chunks_[tid], chunks_[tid+1]);
""",
    'norm2': """#pragma omp master
{
    _norm2_%(var)s = 0.0;
}
auto local_norm2_%(var)s = norm2_value(%(var)s.data(), chunks_[tid], chunks_[tid+1]);
"""
}

global_operation_templates_omp_reduce = {
    'max': """#pragma omp for schedule(static, 1)
for (int t = 0; t < nt; t++)
{
    #pragma omp critical
    {
        if ( local_max_%(var)s > _max_%(var)s )
            _max_%(var)s = local_max_%(var)s;
    }
}
""",
    'min': """#pragma omp for schedule(static, 1)
for (int t = 0; t < nt; t++)
{
    #pragma omp critical
    {
        if ( local_min_%(var)s < _min_%(var)s )
            _min_%(var)s = local_min_%(var)s;
    }
}
""",
    'mean': """#pragma omp for reduction(+: _mean_%(var)s)
for (int t = 0; t < nt; t++)
{
    _mean_%(var)s += local_mean_%(var)s;
}
""",
    'norm1': """#pragma omp for reduction(+: _norm1_%(var)s)
for (int t = 0; t < nt; t++)
{
    _norm1_%(var)s += local_norm1_%(var)s;
}
""",
    'norm2': """#pragma omp for reduction(+: _norm2_%(var)s)
for (int t = 0; t < nt; t++)
{
    _norm2_%(var)s += local_norm2_%(var)s;
}
#pragma omp master
{
    _norm2_%(var)s = sqrt(_norm2_%(var)s);
}
"""
}

global_operation_templates_omp_extern = {
    'max': "%(type)s max_value(const %(type)s*, const int, const int);\n",
    'min': "extern %(type)s min_value(const %(type)s*, const int, const int);\n",
    'mean': "extern %(type)s mean_value(const %(type)s*, const int, const int, const int);\nextern %(type)s mean_value(const %(type)s*, const int, const int);\n",
    'norm1': "extern %(type)s norm1_value(const %(type)s*, const int, const int);\n",
    'norm2': "extern %(type)s norm2_value(const %(type)s*, const int, const int);\n"
}

#
# determine correct kernel sizes
#
global_operation_templates_cuda = {
   'max' : {
        'body' : """// Computes the maximum value of an array
__global__ void cuMaxValue(%(type)s* result, %(type)s *gpu_array, int N)
{
    unsigned int tid = threadIdx.x;
    unsigned int i = tid;

    extern %(type)s __shared__ sdata[];
    %(type)s localMax = FLT_MIN;

    while(i < N)
    {
        localMax = fmaxf(localMax, gpu_array[i]);
        i+= blockDim.x;
    }

    sdata[tid] = localMax;
    __syncthreads();

    // do reduction in shared mem
    if (blockDim.x >= 512) { if (tid < 256) { sdata[tid] = localMax = fmaxf(localMax, sdata[tid + 256]); } __syncthreads(); }
    if (blockDim.x >= 256) { if (tid < 128) { sdata[tid] = localMax = fmaxf(localMax, sdata[tid + 128]); } __syncthreads(); }
    if (blockDim.x >= 128) { if (tid <  64) { sdata[tid] = localMax = fmaxf(localMax, sdata[tid +  64]); } __syncthreads(); }

    if (tid < 32)
    {
        volatile %(type)s* smem = sdata;

        if (blockDim.x >=  64) { smem[tid] = localMax = fmaxf(localMax, smem[tid + 32]); }
        if (blockDim.x >=  32) { smem[tid] = localMax = fmaxf(localMax, smem[tid + 16]); }
        if (blockDim.x >=  16) { smem[tid] = localMax = fmaxf(localMax, smem[tid +  8]); }
        if (blockDim.x >=   8) { smem[tid] = localMax = fmaxf(localMax, smem[tid +  4]); }
        if (blockDim.x >=   4) { smem[tid] = localMax = fmaxf(localMax, smem[tid +  2]); }
        if (blockDim.x >=   2) { smem[tid] = localMax = fmaxf(localMax, smem[tid +  1]); }
    }

    // write result for this block to global mem
    if (tid == 0)
    {
        *result = sdata[0];
    }
}
""",
        'invoke': """void cuda_max_value(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N) {
    cuMaxValue <<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>> ( result, gpu_array, N );
}
""",
        'header': """void cuda_max_value(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N);""",

        'call' : """
    if ( pop%(id)s->_active ) {
        cuda_max_value(RunConfig(1, 32, 64 * sizeof(%(type)s), pop%(id)s->stream), pop%(id)s->_gpu_%(op)s_%(var)s, pop%(id)s->gpu_%(var)s, pop%(id)s->size );
        cudaMemcpy(&pop%(id)s->_max_%(var)s, pop%(id)s->_gpu_%(op)s_%(var)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
    #ifdef _DEBUG
        auto glob_%(op)s_pop%(id)s_err = cudaGetLastError();
        if (glob_%(op)s_pop%(id)s_err != cudaSuccess)
            std::cerr << "Global operation '%(op)s' (PopStruct%(id)s): " <<  << std::endl;
    #endif
    }
"""
    },

    'min' : {
        'body' : """// Computes the minimum value of an array
__global__ void cuMinValue(%(type)s* result, %(type)s *gpu_array, int N)
{
    unsigned int tid = threadIdx.x;
    unsigned int i = tid;

    extern %(type)s __shared__ sdata[];
    %(type)s localMin = FLT_MAX;

    while(i < N)
    {
        localMin = fminf(localMin, gpu_array[i]);
        i+= blockDim.x;
    }

    sdata[tid] = localMin;
    __syncthreads();

    // do reduction in shared mem
    if (blockDim.x >= 512) { if (tid < 256) { sdata[tid] = localMin = fminf(localMin, sdata[tid + 256]); } __syncthreads(); }
    if (blockDim.x >= 256) { if (tid < 128) { sdata[tid] = localMin = fminf(localMin, sdata[tid + 128]); } __syncthreads(); }
    if (blockDim.x >= 128) { if (tid <  64) { sdata[tid] = localMin = fminf(localMin, sdata[tid +  64]); } __syncthreads(); }

    if (tid < 32)
    {
        volatile %(type)s* smem = sdata;

        if (blockDim.x >=  64) { smem[tid] = localMin = fminf(localMin, smem[tid + 32]); }
        if (blockDim.x >=  32) { smem[tid] = localMin = fminf(localMin, smem[tid + 16]); }
        if (blockDim.x >=  16) { smem[tid] = localMin = fminf(localMin, smem[tid +  8]); }
        if (blockDim.x >=   8) { smem[tid] = localMin = fminf(localMin, smem[tid +  4]); }
        if (blockDim.x >=   4) { smem[tid] = localMin = fminf(localMin, smem[tid +  2]); }
        if (blockDim.x >=   2) { smem[tid] = localMin = fminf(localMin, smem[tid +  1]); }
    }

    // write result for this block to global mem
    if (tid == 0)
    {
        *result = sdata[0];
    }
}
""",
        'invoke' : """
void cuda_min_value(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N) {
    cuMinValue <<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>> ( result, gpu_array, N );
}
""",
        'header' : """void cuda_min_value(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N);""",
        'call' : """
    if ( pop%(id)s->_active ) {
        cuda_min_value(RunConfig(1, 32, 64 * sizeof(%(type)s), pop%(id)s->stream), pop%(id)s->_gpu_%(op)s_%(var)s, pop%(id)s->gpu_%(var)s, pop%(id)s->size );
        cudaMemcpy(&pop%(id)s->_min_%(var)s, pop%(id)s->_gpu_%(op)s_%(var)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
    }
"""
    },

    'mean' : {
        'body' : """// Computes the mean value of an array
__global__ void cuMeanValue(%(type)s* result, %(type)s *gpu_array, int N)
{
    unsigned int tid = threadIdx.x;
    unsigned int i = tid;

    extern %(type)s __shared__ sdata[];
    %(type)s localSum = 0.0;

    while(i < N)
    {
        localSum += gpu_array[i];
        i+= blockDim.x;
    }

    sdata[tid] = localSum;
    __syncthreads();

    // do reduction in shared mem
    if (blockDim.x >= 512) { if (tid < 256) { sdata[tid] = localSum = localSum + sdata[tid + 256]; } __syncthreads(); }
    if (blockDim.x >= 256) { if (tid < 128) { sdata[tid] = localSum = localSum + sdata[tid + 128]; } __syncthreads(); }
    if (blockDim.x >= 128) { if (tid <  64) { sdata[tid] = localSum = localSum + sdata[tid +  64]; } __syncthreads(); }

    if (tid < 32)
    {
        volatile %(type)s* smem = sdata;

        if (blockDim.x >=  64) { smem[tid] = localSum = localSum + smem[tid + 32]; }
        if (blockDim.x >=  32) { smem[tid] = localSum = localSum + smem[tid + 16]; }
        if (blockDim.x >=  16) { smem[tid] = localSum = localSum + smem[tid +  8]; }
        if (blockDim.x >=   8) { smem[tid] = localSum = localSum + smem[tid +  4]; }
        if (blockDim.x >=   4) { smem[tid] = localSum = localSum + smem[tid +  2]; }
        if (blockDim.x >=   2) { smem[tid] = localSum = localSum + smem[tid +  1]; }
    }

    // write result for this block to global mem
    if (tid == 0)
    {
        *result = sdata[0] / (%(type)s)N;
    }
}
""",
        'invoke' : """void cuda_mean_value(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N) {
    cuMeanValue <<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>> ( result, gpu_array, N );
}
""",
        'header' : """void cuda_mean_value(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N);""",
        'call' : """
    if ( pop%(id)s->_active ) {
        cuda_mean_value(RunConfig(1, 32, 64 * sizeof(%(type)s), pop%(id)s->stream), pop%(id)s->_gpu_%(op)s_%(var)s, pop%(id)s->gpu_%(var)s, pop%(id)s->size );
        cudaMemcpy(&pop%(id)s->_%(op)s_%(var)s, pop%(id)s->_gpu_%(op)s_%(var)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
    }
"""
    },

    # Computes the L1-norm of an array
    'norm1' : {
        'body' : """// Computes the L1-norm value of an array
__global__ void cuNorm1(%(type)s* result, %(type)s *gpu_array, int N)
{
    unsigned int tid = threadIdx.x;
    unsigned int i = tid;

    extern %(type)s __shared__ sdata[];
    %(type)s localSum = 0.0;

    while(i < N)
    {
        localSum += fabs(gpu_array[i]);
        i+= blockDim.x;
    }

    sdata[tid] = localSum;
    __syncthreads();

    // do reduction in shared mem
    if (blockDim.x >= 512) { if (tid < 256) { sdata[tid] = localSum = localSum + sdata[tid + 256]; } __syncthreads(); }
    if (blockDim.x >= 256) { if (tid < 128) { sdata[tid] = localSum = localSum + sdata[tid + 128]; } __syncthreads(); }
    if (blockDim.x >= 128) { if (tid <  64) { sdata[tid] = localSum = localSum + sdata[tid +  64]; } __syncthreads(); }

    if (tid < 32)
    {
        volatile %(type)s* smem = sdata;

        if (blockDim.x >=  64) { smem[tid] = localSum = localSum + smem[tid + 32]; }
        if (blockDim.x >=  32) { smem[tid] = localSum = localSum + smem[tid + 16]; }
        if (blockDim.x >=  16) { smem[tid] = localSum = localSum + smem[tid +  8]; }
        if (blockDim.x >=   8) { smem[tid] = localSum = localSum + smem[tid +  4]; }
        if (blockDim.x >=   4) { smem[tid] = localSum = localSum + smem[tid +  2]; }
        if (blockDim.x >=   2) { smem[tid] = localSum = localSum + smem[tid +  1]; }
    }

    // write back result
    if (tid == 0)
    {
        *result = sdata[0];
    }
}
""",
        'invoke': """
void cuda_norm1(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N) {
    cuNorm1<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(result, gpu_array, N);
}
""",
        'header' : """void cuda_norm1(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N);""",
        'call' : """
    if ( pop%(id)s->_active ) {
        cuda_norm1(RunConfig(1, 32, 64 * sizeof(%(type)s), pop%(id)s->stream), pop%(id)s->_gpu_%(op)s_%(var)s, pop%(id)s->gpu_%(var)s, pop%(id)s->size );
        cudaMemcpy(&pop%(id)s->_%(op)s_%(var)s, pop%(id)s->_gpu_%(op)s_%(var)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
    }
"""
    },

    # Computes the L2-norm (Euclidian) of an array
    'norm2' : {
        'body' : """// Computes the L2-norm value of an array
__global__ void cuNorm2(%(type)s* result, %(type)s *gpu_array, int N)
{
    unsigned int tid = threadIdx.x;
    unsigned int i = tid;

    extern %(type)s __shared__ sdata[];
    %(type)s localSum = 0.0;

    while(i < N)
    {
        localSum += gpu_array[i]*gpu_array[i];
        i+= blockDim.x;
    }

    sdata[tid] = localSum;
    __syncthreads();

    // do reduction in shared mem
    if (blockDim.x >= 512) { if (tid < 256) { sdata[tid] = localSum = localSum + sdata[tid + 256]; } __syncthreads(); }
    if (blockDim.x >= 256) { if (tid < 128) { sdata[tid] = localSum = localSum + sdata[tid + 128]; } __syncthreads(); }
    if (blockDim.x >= 128) { if (tid <  64) { sdata[tid] = localSum = localSum + sdata[tid +  64]; } __syncthreads(); }

    if (tid < 32)
    {
        volatile %(type)s* smem = sdata;

        if (blockDim.x >=  64) { smem[tid] = localSum = localSum + smem[tid + 32]; }
        if (blockDim.x >=  32) { smem[tid] = localSum = localSum + smem[tid + 16]; }
        if (blockDim.x >=  16) { smem[tid] = localSum = localSum + smem[tid +  8]; }
        if (blockDim.x >=   8) { smem[tid] = localSum = localSum + smem[tid +  4]; }
        if (blockDim.x >=   4) { smem[tid] = localSum = localSum + smem[tid +  2]; }
        if (blockDim.x >=   2) { smem[tid] = localSum = localSum + smem[tid +  1]; }
    }

    // write back result
    if (tid == 0)
    {
        *result = sqrt(sdata[0]);
    }
}
""",
        'invoke': """
void cuda_norm2(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N) {
    cuNorm2<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(result, gpu_array, N);
}
""",
        'header' : """void cuda_norm2(RunConfig cfg, %(type)s* result, %(type)s *gpu_array, int N);""",
        'call' : """
    if ( pop%(id)s->_active ) {
        cuda_norm2(RunConfig(1, 32, 64 * sizeof(%(type)s), pop%(id)s->stream), pop%(id)s->_gpu_%(op)s_%(var)s, pop%(id)s->gpu_%(var)s, pop%(id)s->size );
        cudaMemcpy(&pop%(id)s->_%(op)s_%(var)s, pop%(id)s->_gpu_%(op)s_%(var)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
    }
"""
    }
}



from .Neurons import list_standard_neurons
from .Neurons import LeakyIntegrator, Izhikevich, IF_curr_exp, IF_cond_exp, IF_curr_alpha, IF_cond_alpha, HH_cond_exp, EIF_cond_alpha_isfa_ista, EIF_cond_exp_isfa_ista
from .Synapses import list_standard_synapses, DefaultRateCodedSynapse, DefaultSpikingSynapse
from .Synapses import STP, STDP, Hebb, Oja, IBCM


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Definition of a population as a c-like struct, divided
# into two groups: rate or spike
#
# Parameters:
#
#    id: id of the population
#    additional: neuron specific definitions
#    accessors: set of functions to export population data to python
population_header = """/*
 *  ANNarchy-version: %(annarchy_version)s
 */
#pragma once
#include "ANNarchy.hpp"
#include <random>
#include "randutils.hpp"
%(include_additional)s
%(include_profile)s
extern %(float_prec)s dt;
extern long int t;
extern int global_num_threads;
extern std::vector<std::mt19937> rng;
%(extern_global_operations)s
%(struct_additional)s
///////////////////////////////////////////////////////////////
// Main Structure for the population of id %(id)s (%(name)s)
///////////////////////////////////////////////////////////////
extern struct PopStruct%(id)s *pop%(id)s;
struct PopStruct%(id)s{

    PopStruct%(id)s(int size, int max_delay) {
        this->size = size;
        this->max_delay = max_delay;

        // HACK: the object constructor is now called by nanobind, need to update reference in C++ library
        pop%(id)s = this;

    #ifdef _TRACE_INIT
        std::cout << "  PopStruct%(id)s - this = " << this << " has been allocated." << std::endl;
    #endif
    }

    int size; // Number of neurons
    bool _active; // Allows to shut down the whole population
    int max_delay; // Maximum number of steps to store for delayed synaptic transmission

    // Access functions used by cython wrapper
    int get_size() { return size; }
    void set_size(int s) { size  = s; }
    int get_max_delay() { return max_delay; }
    void set_max_delay(int d) { max_delay  = d; }
    bool is_active() { return _active; }
    void set_active(bool val) { _active = val; }

%(declare_spike_arrays)s
    // Neuron specific parameters and variables
%(declare_parameters_variables)s
%(declare_delay)s
%(declare_FR)s
%(declare_additional)s
%(declare_profile)s
    // Access methods to the parameters and variables
%(access_parameters_variables)s
%(access_additional)s

    // Method called to initialize the data structures
    void init_population() {
    #ifdef _TRACE_INIT
        std::cout << "  PopStruct%(id)s::init_population(size="<<this->size<<") - this = " << this << std::endl;
    #endif    
        _active = true;
%(init_parameters_variables)s
%(init_spike)s
%(init_delay)s
%(init_FR)s
%(init_additional)s
%(init_profile)s
    }

    // Method called to reset the population
    void reset() {
%(reset_spike)s
%(reset_delay)s
%(reset_additional)s
    }

    // Method to draw new random numbers
    void update_rng(int tid) {
    #ifdef _TRACE_SIMULATION_STEPS
        #pragma omp critical
        {
            std::cout << "    PopStruct%(id)s::update_rng() - tid " << tid << std::endl;
            std::cout << std::flush;
        }
    #endif
%(update_rng)s
    }

    // Method to update global operations on the population (min/max/mean...)
    void update_global_ops(int tid, int nt) {
%(update_global_ops)s
    }

    // Method to enqueue output variables in case outgoing projections have non-zero delay
    void update_delay() {
%(update_delay)s
    }

    // Method to dynamically change the size of the queue for delayed variables
    void update_max_delay(int value) {
%(update_max_delay)s
    }

    // Main method to update neural variables
    void update(int tid) {
    #ifdef _TRACE_SIMULATION_STEPS
        #pragma omp critical
        {
            std::cout << "    PopStruct%(id)s::update() - tid " << tid << std::endl;
            std::cout << std::flush;
        }
    #endif
%(update_variables)s
    }

    void spike_gather(int tid) {
    #ifdef _TRACE_SIMULATION_STEPS
        #pragma omp critical
        {
            std::cout << "    PopStruct%(id)s::spike_gather() - tid " << tid << std::endl;
            std::cout << std::flush;
        }
    #endif
%(test_spike_cond)s
    }

    %(stop_condition)s

    // Memory management: track the memory consumption
    long int size_in_bytes() {
        long int size_in_bytes = 0;
%(size_in_bytes)s
        return size_in_bytes;
    }

    // Memory management: track the memory consumption
    void clear() {
%(clear_container)s
    }
};
"""

# c like definition of neuron attributes, whereas 'local' is used if values can vary across
# neurons, consequently 'global' is used if values are common to all neurons.Currently two
# types of sets are defined: openmp and cuda. In cuda case additional 'dirty' flags are
# created.
#
# Parameters:
#
#    type: data type of the variable (double, float, int ...)
#    name: name of the variable
#    attr_type: either 'variable' or 'parameter'
#
attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

# Initialization of parameters due to the init_population method.
#
# Parameters:
#
#    name: name of the variable
#    init: initial value
attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = std::vector<%(type)s>(size, %(init)s);
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_delayed = {
    'local': {
        'init': """
        _delayed_%(name)s = std::deque< std::vector< %(type)s > >(max_delay, std::vector< %(type)s >(size, 0.0));""",

        'update': """
        #pragma omp single
        {
            _delayed_%(name)s.push_front(%(name)s);
            _delayed_%(name)s.pop_back();
        }
""",
        'reset' : """
        for ( int i = 0; i < _delayed_%(name)s.size(); i++ ) {
            _delayed_%(name)s[i] = %(name)s;
        }
""",
        'resize' : """
    _delayed_%(name)s.resize(max_delay, std::vector< %(type)s >(size, 0.0));
"""
    },
    'global':{
        'init': """
        _delayed_%(name)s = std::deque< %(type)s >(max_delay, 0.0);""",
        'update': """
        #pragma omp single
        {
            _delayed_%(name)s.push_front(%(name)s);
            _delayed_%(name)s.pop_back();
        }
""",
        'reset' : """
        for ( int i = 0; i < _delayed_%(name)s.size(); i++ ) {
            _delayed_%(name)s[i] = %(name)s;
        }
""",
        'resize' : """
    _delayed_%(name)s.resize(max_delay, 0.0);
"""
    }
}

# Definition for the usage of C++11 STL template random
# number generators
#
# Parameters:
#
#    rd_name:
#    rd_update:
cpp_11_rng = {
    'dist_decl': "auto dist_%(rd_name)s = %(rd_init)s;",
    'local': {
        'decl': "std::vector<%(type)s> %(rd_name)s ;",
        'init': "%(rd_name)s = std::vector<%(type)s>(size, 0.0);",
        'update': "%(rd_name)s[i] = dist_%(rd_name)s(rng[%(index)s]);",
        'clear': "%(rd_name)s.clear();\n%(rd_name)s.shrink_to_fit();"
    },
    'global': {
        'decl': "%(type)s %(rd_name)s;",
        'init': "%(rd_name)s = 0.0;",
        'update': "%(rd_name)s = dist_%(rd_name)s(rng[0]);",
        'clear': ""
    },
    'omp_code_seq': """
        if (_active){

            #pragma omp single
            {
%(rng_dist)s
%(update_rng_global)s
%(update_rng_local)s
            }
        }
    """,
    'omp_code_par': """
        if (_active){
%(rng_dist)s
%(update_rng_global)s
%(update_rng_local)s
        }
    """
}

rate_psp = {
    'decl': """
    std::vector<%(float_prec)s> _sum_%(target)s;""",
    'init': """
        // Post-synaptic potential
        _sum_%(target)s = std::vector<%(float_prec)s>(size, 0.0);""",
    'reset': """
    // pop%(id)s: %(name)s
    #pragma omp single nowait
    {
        if (pop%(id)s->_active)
            std::fill(pop%(id)s->_sum_%(target)s.begin(), pop%(id)s->_sum_%(target)s.end(), static_cast<%(float_prec)s>(0.0));
    }
"""
}

spike_specific = {
    'spike': {
        'declare': """
    // Structures for managing spikes
    std::vector<long int> last_spike;
    std::vector<int> spiked;
    std::vector<int> local_spiked_sizes;
""",
        'init': """
        // Spiking variables
        spiked = std::vector<int>();
        local_spiked_sizes = std::vector<int>(global_num_threads+1, 0);
        last_spike = std::vector<long int>(size, -10000L);
""",
        'reset': """
        spiked.clear();
        spiked.shrink_to_fit();
        local_spiked_sizes = std::vector<int>(global_num_threads+1, 0);
        last_spike.clear();
        last_spike = std::vector<long int>(size, -10000L);
""",
        'clear': """
// Spike events
spiked.clear();
spiked.shrink_to_fit();

last_spike.clear();
last_spike.shrink_to_fit();
"""
    },
    'axon_spike': {
        'declare': """
    // Structures for managing axonal spikes
    std::vector<int> axonal;
""",
        'init': """
        // Axonal spike containter
        axonal = std::vector<int>();
""",
        'reset': """
        axonal.clear();
        axonal.shrink_to_fit();
""",
        'pyx_wrapper': """
    # Axonal spike events
"""
    },
    'refractory': {

        'declare': """
    // Refractory period
    std::vector<int> refractory;
    std::vector<int> refractory_remaining;
    std::vector<short int> in_ref;
""",

        'init': """
        // Refractory period
        refractory = std::vector<int>(size, 0);
        refractory_remaining = std::vector<int>(size, 0);
        in_ref = std::vector<short int>(size, 0);
""",

        # If the refractory variable is defined by the user
        'init_extern': """
        // Refractory period
        refractory_remaining = std::vector<int>(size, 0);
        in_ref = std::vector<short int>(size, 0);
""",

        'reset': """
        // Refractory period
        refractory_remaining.clear();
        refractory_remaining = std::vector<int>(size, 0);
"""
    }
}

#
# Final dictionary
openmp_templates = {
    'population_header': population_header,
    'attr_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_delayed': attribute_delayed,
    'rng': cpp_11_rng,

    'rate_psp': rate_psp,
    'spike_specific': spike_specific
}



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Synapse import Synapse
from ANNarchy.intern import Messages


def list_standard_synapses():
    "Returns a list of standard neuron models available."
    return [STP, STDP, Hebb, Oja, IBCM]


###############################
### Default rate-coded Synapse
###############################
class DefaultRateCodedSynapse(Synapse):
    # For reporting
    _instantiated = []
    def __init__(self):
        Synapse.__init__(self, 
            equations="",
            psp="w * pre.r",
            name="-", 
            description="Weighted sum of firing rates."
        )
        # For reporting
        self._instantiated.append(True)

###############################
### Default spiking Synapse
###############################
class DefaultSpikingSynapse(Synapse):
    # For reporting
    _instantiated = []
    def __init__(self):
        Synapse.__init__(self, 
            pre_spike = "g_target += w",
            name="Event-driven synapse", 
            description="Increases the post-synaptic conductance from the synaptic efficency after each pre-synaptic spike."
        )
        # For reporting
        self._instantiated.append(True)

##################
### Hebb
##################
class Hebb(Synapse):
    '''
    Rate-coded synapse with Hebbian plasticity.

    Equivalent code:

    ```python
    Hebb = ann.Synapse(
        parameters = dict(
            eta = 0.01,
        ),
        equations = [
            ann.Variable('dw/dt = eta * pre.r * post.r', min=0.0),
        ]
    )
    ```

    :param eta: learning rate.
    '''
    # For reporting
    _instantiated = []

    def __init__(self, eta=0.01):

        parameters = """
            eta = %(eta)s : projection
        """ % {'eta': eta}

        equations = """
            dw/dt = eta * pre.r * post.r : min=0.0, explicit 
        """

        Synapse.__init__(
            self, 
            parameters=parameters, 
            equations=equations,
            name="Hebbian Plasticity", 
            description="Simple Hebbian learning rule"
        )
        
        # For reporting
        self._instantiated.append(True)

##################
### Oja
##################
class Oja(Synapse):
    '''
    Rate-coded synapse with regularized Hebbian plasticity (Oja).
    
    Equivalent code:

    ```python
    Oja = ann.Synapse(
        parameters = dict(
            eta = 0.01,
            alpha = 1.0,
        ),
        equations = [
            ann.Variable('dw/dt = eta * ( pre.r * post.r - alpha * post.r^2 * w )', min=0.0),
        ]
    )
    ```

    :param eta: learning rate.
    :param alpha: regularization coefficient.
    '''
    # For reporting
    _instantiated = []

    def __init__(self, eta:float=0.01, alpha:float=1.0):

        parameters = """
            eta = %(eta)s : projection
            alpha = %(alpha)s : projection
        """ % {'eta': eta, 'alpha': alpha}

        equations = """
            dw/dt = eta * ( pre.r * post.r - alpha * post.r^2 * w ) : min=0.0, explicit 
        """

        Synapse.__init__(self, parameters=parameters, equations=equations,
            name="Oja plasticity", description="Regularized Hebbian learning rule.")
        # For reporting
        self._instantiated.append(True)

##################
### IBCM
##################
class IBCM(Synapse):
    '''
    Rate-coded synapse with Intrator & Cooper (1992) plasticity.

    Equivalent code:

    ```python
    IBCM = ann.Synapse(
        parameters = dict(
            eta = 0.01,
            tau = 2000.0,
        ),
        equations = [
            ann.Variable('tau * dtheta/dt + theta = post.r^2', locality='semiglobal', method='exponential'),
            ann.Variable('dw/dt = eta * post.r * (post.r - theta) * pre.r',' min=0.0),
        ]
    )
    ```

    :param eta: learning rate.
    :param tau: time constant of the sliding threshold.
    '''
    # For reporting
    _instantiated = []

    def __init__(self, eta = 0.01, tau = 2000.0):

        parameters = """
            eta = %(eta)s : projection
            tau = %(tau)s : projection
        """ % {'eta': eta, 'tau': tau}

        equations = """
            tau * dtheta/dt + theta = post.r^2 : postsynaptic, exponential
            dw/dt = eta * post.r * (post.r - theta) * pre.r : min=0.0, explicit
        """

        Synapse.__init__(self, parameters=parameters, equations=equations,
            name="IBCM", description="Intrator and Cooper (1992) learning rule.")
        # For reporting
        self._instantiated.append(True)

##################
### STP
##################
class STP(Synapse):
    '''
    Synapse exhibiting short-term facilitation and depression.
    
    Implemented using the model of Tsodyks, Markram et al.:

    > Tsodyks, Uziel and Markram (2000) Synchrony Generation in Recurrent Networks with Frequency-Dependent Synapses. Journal of Neuroscience 20:RC50

    Note that the time constant of the post-synaptic current is set in the neuron model, not here.
    
    Equivalent code:

    ```python
    STP = ann.Synapse(
        parameters = dict(
            tau_rec = 100.0,
            tau_facil = 0.01,
            U = 0.5,
        ),
        equations = [
            ann.Variable('dx/dt = (1 - x)/tau_rec', init = 1.0, method='event-driven'),
            ann.Variable('du/dt = (U - u)/tau_facil', init = 0.5, method='event-driven'),
        ],
        pre_spike="""
            g_target += w * u * x
            x *= (1 - u)
            u += U * (1 - u)
        """
    )
    ```

    :param tau_rec: depression time constant (ms).
    :param tau_facil: facilitation time constant (ms).
    :param U: use parameter.
    '''
    # For reporting
    _instantiated = []

    def __init__(self, tau_rec=100.0, tau_facil=0.01, U=0.5):

        if tau_facil<= 0.0:
            Messages._error('STP: tau_facil must be positive. Choose a very small value if you have to, or derive a new synapse.')
            
        parameters = """
            tau_rec = %(tau_rec)s : projection
            tau_facil = %(tau_facil)s : projection
            U = %(U)s
        """ % {'tau_rec': tau_rec, 'tau_facil': tau_facil, 'U': U}
        equations = """
            dx/dt = (1 - x)/tau_rec : init = 1.0, event-driven
            du/dt = (U - u)/tau_facil : init = %(U)s, event-driven   
        """ % {'tau_rec': tau_rec, 'tau_facil': tau_facil, 'U': U}
        pre_spike="""
            g_target += w * u * x
            x *= (1 - u)
            u += U * (1 - u)
        """

        Synapse.__init__(self, parameters=parameters, equations=equations, pre_spike=pre_spike,
            name="Short-term plasticity", description="Synapse exhibiting short-term facilitation and depression, implemented using the model of Tsodyks, Markram et al.")
        # For reporting
        self._instantiated.append(True)
        

##################
### STDP
##################
class STDP(Synapse):
    '''
    Spike-timing dependent plasticity, online version.

    > Song, S., and Abbott, L.F. (2001). Cortical development and remapping through spike timing-dependent plasticity. Neuron 32, 339-350. 
    
    Equivalent code:

    ```python
    STDP = ann.Synapse(
        parameters = dict(
            tau_plus = 20.0,
            tau_minus = 20.0,
            A_plus = 0.01,
            A_minus = 0.01,
            w_min = 0.0,
            w_max = 1.0,
        ),
        equations = [
            ann.Variable('tau_plus  * dx/dt = -x', method='event-driven'),
            ann.Variable('tau_minus * dy/dt = -y', method='event-driven'),
        ],
        pre_spike="""
            g_target += w
            x += A_plus * w_max
            w = clip(w + y, w_min , w_max)
        """,
        post_spike="""
            y -= A_minus * w_max
            w = clip(w + x, w_min , w_max)
        """
    )
    ```

    
    :param tau_plus: time constant of the pre-synaptic trace (ms)
    :param tau_minus: time constant of the pre-synaptic trace (ms)
    :param A_plus: increase of the pre-synaptic trace after a spike.
    :param A_minus: decrease of the post-synaptic trace after a spike. 
    :param w_min: minimal value of the weight w.
    :param w_max: maximal value of the weight w.
    '''
    # For reporting
    _instantiated = []

    def __init__(self, tau_plus=20.0, tau_minus=20.0, A_plus=0.01, A_minus=0.01, w_min=0.0, w_max=1.0):

        parameters="""
            tau_plus = %(tau_plus)s : projection
            tau_minus = %(tau_minus)s : projection
            A_plus = %(A_plus)s : projection
            A_minus = %(A_minus)s : projection
            w_min = %(w_min)s : projection
            w_max = %(w_max)s : projection
        """ % {'tau_plus': tau_plus, 'tau_minus':tau_minus, 'A_plus':A_plus, 'A_minus': A_minus, 'w_min': w_min, 'w_max': w_max}

        equations = """
            tau_plus  * dx/dt = -x : event-driven
            tau_minus * dy/dt = -y : event-driven
        """
        pre_spike="""
            g_target += w
            x += A_plus * w_max
            w = clip(w + y, w_min , w_max)
        """          
        post_spike="""
            y -= A_minus * w_max
            w = clip(w + x, w_min , w_max)
        """

        Synapse.__init__(self, parameters=parameters, equations=equations, pre_spike=pre_spike, post_spike=post_spike,
            name="Spike-timing dependent plasticity", description="Synapse exhibiting spike-timing dependent plasticity.")
        # For reporting
        self._instantiated.append(True)





"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global %(attr_type)s %(name)s
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

delay = {
    # A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
""",
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': None,
    # An individual value for each synapse and a
    # buffer for spike events    
    'nonuniform_spiking': None
}

event_driven = {
    'declare': """
    std::vector<long> _last_event;
""",
    'cpp_init': """
        // Event-driven
        _last_event = init_matrix_variable<long>(-10000);
""",
    'pyx_struct': """
        vector[vector[long]] _last_event
""",
}

######################################
### Dense Matrix templates
######################################
dense_summation_operation = {
    'sum' : """
%(pre_copy)s

// matrix dimensions
%(idx_type)s rows = %(post_prefix)ssize;
%(idx_type)s columns = %(pre_prefix)ssize;

// running indices
%(idx_type)s i;
%(size_type)s j;

// target array
%(float_prec)s* __restrict__ target = %(post_prefix)s_sum_%(target)s.data();

%(omp_code)s %(omp_clause)s %(omp_schedule)s
for(%(idx_type)s _idx = 0; _idx < post_ranks_.size(); _idx++) {
    i = post_ranks_[_idx];

    %(float_prec)s sum = 0.0;
    for(%(idx_type)s rk_pre = 0, j=i*columns; rk_pre < columns; j++, rk_pre++) {
        sum += %(psp)s ;
    }
    target[i] += sum;
}
"""
}

###############################################################################
# Optimized kernel for default rate-coded continuous transmission using 
# SIMD intrinsics (SEE4-1, AVX)
###############################################################################
continuous_transmission_sse = {
    'sum': {
        'double': """
#ifdef __SSE4_1__
    if (_transmission && pop%(id_post)s->_active) {
        double _tmp_sum[2];

        // matrix dimensions
        %(idx_type)s rows = pop%(id_post)s->size;
        %(idx_type)s columns = pop%(id_pre)s->size;

        // running indices
        %(idx_type)s i, j;
        %(size_type)s _s;

        // required pointer
        double* __restrict__ _pre_r = %(get_r)s;
        double* __restrict__ _w = w.data();

        // Row-wise SpMV
        #pragma omp for
        for(%(idx_type)s _idx = 0; _idx < post_ranks_.size(); _idx++) {
            i = post_ranks_[_idx];

            __m128d _tmp_reg_sum = _mm_setzero_pd();

            _s=i*columns;
            for (j = 0; (j+8) < columns; j+=8, _s+=8) {
                __m128d _tmp_r = _mm_loadu_pd(&_pre_r[j]);
                __m128d _tmp_r2 = _mm_loadu_pd(&_pre_r[j+2]);
                __m128d _tmp_r3 = _mm_loadu_pd(&_pre_r[j+4]);
                __m128d _tmp_r4 = _mm_loadu_pd(&_pre_r[j+6]);

                __m128d _tmp_w = _mm_loadu_pd(&_w[_s]);
                __m128d _tmp_w2 = _mm_loadu_pd(&_w[_s+2]);
                __m128d _tmp_w3 = _mm_loadu_pd(&_w[_s+4]);
                __m128d _tmp_w4 = _mm_loadu_pd(&_w[_s+6]);

                _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r, _tmp_w));
                _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r2, _tmp_w2));
                _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r3, _tmp_w3));
                _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r4, _tmp_w4));
            }
            _mm_storeu_pd(_tmp_sum, _tmp_reg_sum);

            // partial sums
            double lsum = _tmp_sum[0] + _tmp_sum[1];

            // remainder loop
            for (; j < columns; j++, _s++)
                lsum += _pre_r[j] * _w[_s];

            pop%(id_post)s._sum_%(target)s[i] += lsum;
        }
    } // active
#else
    std::cerr << "The code was not compiled with SSE4-1 support. Please check your compiler flags ..." << std::endl;
#endif
""",
        'float': """
#ifdef __SSE4_1__
    if (_transmission && pop%(id_post)s->_active) {
        float _tmp_sum[4];

        // matrix dimensions
        %(idx_type)s rows = pop%(id_post)s->size;
        %(idx_type)s columns = pop%(id_pre)s->size;

        // running indices
        %(idx_type)s i, j;
        %(size_type)s _s;

        // required pointer
        float* __restrict__ _pre_r = %(get_r)s;
        float* __restrict__ _w = w.data();

        // Row-wise SpMV
        #pragma omp for
        for(%(idx_type)s _idx = 0; _idx < post_ranks_.size(); _idx++) {
            i = post_ranks_[_idx];

            __m128 _tmp_reg_sum = _mm_setzero_ps();

            _s=i*columns;
            for (j = 0; (j+16) < columns; j+=16, _s+=16) {
                __m128 _tmp_r = _mm_loadu_ps(&_pre_r[j]);
                __m128 _tmp_r2 = _mm_loadu_ps(&_pre_r[j+4]);
                __m128 _tmp_r3 = _mm_loadu_ps(&_pre_r[j+8]);
                __m128 _tmp_r4 = _mm_loadu_ps(&_pre_r[j+12]);

                __m128 _tmp_w = _mm_loadu_ps(&_w[_s]);
                __m128 _tmp_w2 = _mm_loadu_ps(&_w[_s+4]);
                __m128 _tmp_w3 = _mm_loadu_ps(&_w[_s+8]);
                __m128 _tmp_w4 = _mm_loadu_ps(&_w[_s+12]);
                
                _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r, _tmp_w));
                _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r2, _tmp_w2));
                _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r3, _tmp_w3));
                _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r4, _tmp_w4));
            }
            _mm_storeu_ps(_tmp_sum, _tmp_reg_sum);

            // partial sums
            float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

            // remainder loop
            for (; j < columns; j++, _s++)
                lsum += _pre_r[j] * _w[_s];

            pop%(id_post)s._sum_%(target)s%(post_index)s += lsum;
        }
    } // active
#else
    std::cerr << "The code was not compiled with SSE4-1 support. Please check your compiler flags ..." << std::endl;
#endif
"""
    }
}

continuous_transmission_avx = {
    'sum': {
        'double': """
#ifdef __AVX__
    if (_transmission && %(post_prefix)s_active) {
        double _tmp_sum[4];

        // matrix dimensions
        %(idx_type)s rows = %(post_prefix)ssize;
        %(idx_type)s columns = %(pre_prefix)ssize;

        // running indices
        %(idx_type)s i, j;
        %(size_type)s _s;

        // required pointer
        double* __restrict__ _pre_r = %(get_r)s;
        double* __restrict__ _w = w.data();

        // Row-wise SpMV
        #pragma omp for
        for(%(idx_type)s _idx = 0; _idx < post_ranks_.size(); _idx++) {
            i = post_ranks_[_idx];

            __m256d _tmp_reg_sum = _mm256_setzero_pd();

            _s=i*columns;
            for (j = 0; (j+8) < columns; j+=8, _s+=8) {
                __m256d _tmp_r = _mm256_loadu_pd(&_pre_r[j]);
                __m256d _tmp_r2 = _mm256_loadu_pd(&_pre_r[j+4]);

                __m256d _tmp_w = _mm256_loadu_pd(&_w[_s]);
                __m256d _tmp_w2 = _mm256_loadu_pd(&_w[_s+4]);

                _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _mm256_mul_pd(_tmp_r, _tmp_w));
                _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _mm256_mul_pd(_tmp_r2, _tmp_w2));
            }
            _mm256_storeu_pd(_tmp_sum, _tmp_reg_sum);

            // partial sums
            double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

            // remainder loop
            for (; j < columns; j++, _s++)
                lsum += _pre_r[j] * _w[_s];

            %(post_prefix)s_sum_%(target)s[i] += lsum;
        }
    } // active
#else
    std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
#endif
""",
    'float': """
    #ifdef __AVX__
    if (_transmission && %(post_prefix)s_active) {
        float _tmp_sum[8];

        // matrix dimensions
        %(idx_type)s rows = %(post_prefix)ssize;
        %(idx_type)s columns = %(pre_prefix)ssize;
        // running indices
        %(idx_type)s i, j;
        %(size_type)s _s;

        // required pointer
        float* __restrict__ _pre_r = %(get_r)s;
        float* __restrict__ _w = w.data();

        // Row-wise SpMV
        #pragma omp for
        for(%(idx_type)s _idx = 0; _idx < post_ranks_.size(); _idx++) {
            i = post_ranks_[_idx];

            __m256 _tmp_reg_sum = _mm256_setzero_ps();

            _s=i*columns;
            for (j = 0; (j+16) < columns; j+=16, _s+=16) {
                __m256 _tmp_r = _mm256_loadu_ps(&_pre_r[j]);
                __m256 _tmp_r2 = _mm256_loadu_ps(&_pre_r[j+8]);

                __m256 _tmp_w = _mm256_loadu_ps(&_w[_s]);
                __m256 _tmp_w2 = _mm256_loadu_ps(&_w[_s+8]);

                _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_mul_ps(_tmp_r, _tmp_w));
                _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_mul_ps(_tmp_r2, _tmp_w2));
            }

            _mm256_storeu_ps(_tmp_sum, _tmp_reg_sum);

            // partial sums
            double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

            // remainder loop
            for (; j < columns; j++, _s++)
                lsum += _pre_r[j] * _w[_s];

            %(post_prefix)s_sum_%(target)s[i] += lsum;
        }
    } // active
#else
    std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
#endif
""",
        'float': """
#ifdef __AVX__
    if (_transmission && pop%(id_post)s->_active) {
        float _tmp_sum[8];

        // matrix dimensions
        %(idx_type)s rows = pop%(id_post)s->size;
        %(idx_type)s columns = pop%(id_pre)s->size;

        // running indices
        %(idx_type)s i, j;
        %(size_type)s _s;

        // required pointer
        float* __restrict__ _pre_r = %(get_r)s;
        float* __restrict__ _w = w.data();

        // Row-wise SpMV
        #pragma omp for
        for(%(idx_type)s _idx = 0; _idx < post_ranks_.size(); _idx++) {
            i = post_ranks_[_idx];

            %(idx_type)s rk_post = i;
            __m256 _tmp_reg_sum = _mm256_setzero_ps();

            _s=i*columns;
            for (j = 0; (j+16) < columns; j+=16, _s+=16) {
                __m256 _tmp_r = _mm256_loadu_ps(&_pre_r[j]);
                __m256 _tmp_r2 = _mm256_loadu_ps(&_pre_r[j+8]);

                __m256 _tmp_w = _mm256_loadu_ps(&_w[_s]);
                __m256 _tmp_w2 = _mm256_loadu_ps(&_w[_s+8]);

                _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_mul_ps(_tmp_r, _tmp_w));
                _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_mul_ps(_tmp_r2, _tmp_w2));
            }
            _mm256_storeu_ps(_tmp_sum, _tmp_reg_sum);

            // partial sums
            float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

            // remainder loop
            for (; j < columns; j++, _s++)
                lsum += _pre_r[j] * _w[_s];

            pop%(id_post)s._sum_%(target)s%(post_index)s += lsum;
        }
    } // active
#else
    std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
#endif
"""
    }
}

continuous_transmission_avx512 = {
    'sum': {
        'double': """
#ifdef __AVX512F__
    if (_transmission && pop%(id_post)s->_active) {
        double _tmp_sum[8];

        // matrix dimensions
        %(idx_type)s rows = pop%(id_post)s->size;
        %(idx_type)s columns = pop%(id_pre)s->size;

        // running indices
        %(idx_type)s i, j;
        %(size_type)s _s;

        // required pointer
        double* __restrict__ _pre_r = %(get_r)s;
        double* __restrict__ _w = w.data();

        // Row-wise SpMV
        #pragma omp for
        for(%(idx_type)s _idx = 0; _idx < post_ranks_.size(); _idx++) {
            i = post_ranks_[_idx];

            %(idx_type)s rk_post = i;
            __m512d _tmp_reg_sum = _mm512_setzero_pd();

            _s=i*columns;
            for (j = 0; (j+8) < columns; j+=8, _s+=8) {
                __m512d _tmp_r = _mm512_loadu_pd(&_pre_r[j]);
                __m512d _tmp_w = _mm512_loadu_pd(&_w[_s]);

                _tmp_reg_sum = _mm512_add_pd(_tmp_reg_sum, _mm512_mul_pd(_tmp_r, _tmp_w));
            }

            _mm512_storeu_pd(_tmp_sum, _tmp_reg_sum);

            // partial sums
            double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

            // remainder loop
            for (; j < columns; j++, _s++)
                lsum += _pre_r[j] * _w[_s];

            pop%(id_post)s._sum_%(target)s%(post_index)s += lsum;
        }
    } // active
#else
    std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
#endif
""",
        'float': """
#ifdef __AVX512F__
    if (_transmission && pop%(id_post)s->_active) {
        float _tmp_sum[16];

        // matrix dimensions
        %(idx_type)s rows = pop%(id_post)s->size;
        %(idx_type)s columns = pop%(id_pre)s->size;

        // running indices
        %(idx_type)s i, j;
        %(size_type)s _s;

        // required pointer
        float* __restrict__ _pre_r = %(get_r)s;
        float* __restrict__ _w = w.data();

        // Row-wise SpMV
        #pragma omp for
        for(%(idx_type)s _idx = 0; _idx < post_ranks_.size(); _idx++) {
            i = post_ranks_[_idx];

            %(idx_type)s rk_post = i;
            __m512 _tmp_reg_sum = _mm512_setzero_ps();

            _s=i*columns;
            for (j = 0; (j+16) < columns; j+=16, _s+=16) {
                __m512 _tmp_r = _mm512_loadu_ps(&_pre_r[j]);
                __m512 _tmp_w = _mm512_loadu_ps(&_w[_s]);

                _tmp_reg_sum = _mm512_add_ps(_tmp_reg_sum, _mm512_mul_ps(_tmp_r, _tmp_w));
            }

            _mm512_storeu_ps(_tmp_sum, _tmp_reg_sum);

            // partial sums
            float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7] + _tmp_sum[8] + _tmp_sum[9] + _tmp_sum[10] + _tmp_sum[11] + _tmp_sum[12] + _tmp_sum[13] + _tmp_sum[14] + _tmp_sum[15];

            // remainder loop
            for (; j < columns; j++, _s++)
                lsum += _pre_r[j] * _w[_s];

            pop%(id_post)s._sum_%(target)s%(post_index)s += lsum;
        }
    } // active
#else
    std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
#endif
"""
    }
}

# HD (23th Aug 2024):
# Each thread computes the effect the pre-synaptic neurons onto his own row.
# Even though the access to post-synaptic variables don't need synchronization,
# the CPU caches will be busy to re-load the target arrays repeatedly ...
spiking_summation_fixed_delay_inner_loop = """// Event-based summation
if (_transmission && %(post_prefix)s_active){

    for (%(idx_type)s rk_post = tid; rk_post < num_rows(); rk_post += nt) {

        // Iterate over all spiking neurons
        for (auto it = %(pre_prefix)sspiked.cbegin(); it != %(pre_prefix)sspiked.cend(); it++) {
            %(size_type)s j = rk_post*this->num_columns_ + (*it);

            if (mask_[j]) {
                // post-synaptic popential
                %(g_target)s

                // 'on-pre' events
                %(event_driven)s
                %(pre_event)s
            }
        }
    }
} // active
"""

spiking_summation_fixed_delay_only_psp_inner_loop = """// Event-based summation
if (_transmission && %(post_prefix)s_active){

    for (%(idx_type)s rk_post = tid; rk_post < num_rows(); rk_post += nt) {

        // Iterate over all spiking neurons
        for (auto it = %(pre_prefix)sspiked.cbegin(); it != %(pre_prefix)sspiked.cend(); it++) {
            %(size_type)s j = rk_post*this->num_columns_ + (*it);
            %(g_target)s
        }
    }
} // active
"""

# HD (19th May 2022):
# Our default strategy, to loop over all spike events and update post.g_target can not applied here
# as it would lead to 100% cache misses and an enormously high number of memory stalls.
spiking_summation_fixed_delay_outer_loop = """// Event-based summation
if (_transmission && %(post_prefix)s_active){

    // Iterate over all spiking neurons
    #pragma omp for
    for (auto it = %(pre_prefix)sspiked.cbegin(); it != %(pre_prefix)sspiked.cend(); it++) {
        %(idx_type)s rk_pre = *it;
        for (%(idx_type)s rk_post = 0; rk_post < num_rows(); rk_post++) {
            %(size_type)s j = rk_post*this->num_columns_ + rk_pre;

            if (mask_[j]) {
                // post-synaptic popential
                #pragma omp atomic%(g_target)s

                // 'on-pre' events
                #pragma omp critical
                {
                %(event_driven)s
                %(pre_event)s
                }
            }
        }
    }
} // active
"""

dense_update_variables = {
    'local': """
// Check periodicity
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L)){
    // Global variables
    %(global)s

    // Local variables
    #pragma omp for
    for(int i = 0; i < %(post_prefix)ssize; i++){
        rk_post = i; // dense: ranks are indices
        // Semi-global variables
    %(semiglobal)s

        // Local variables
        %(size_type)s j = i*%(pre_prefix)ssize;
        for (rk_pre = 0; rk_pre < %(pre_prefix)ssize; rk_pre++, j++) {
            if(mask_[j]) {
    %(local)s
            }
        }
    }
}
""",
    'global': """
// Check periodicity
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L)){
    // Global variables
    %(global)s

    // Semi-global variables
    #pragma omp for
    for(int i = 0; i < %(post_prefix)ssize; i++){
        rk_post = i;
    %(semiglobal)s
    }
}
"""
}

spiking_post_event = """
if (_transmission && %(post_prefix)s_active) {

    %(idx_type)s rows = pop%(id_pre)s->size;
    %(idx_type)s columns = pop%(id_post)s->size;

    for (%(idx_type)s _idx_i = tid; _idx_i < %(post_prefix)sspiked.size(); _idx_i+= nt) {
        %(idx_type)s post_rank = %(post_prefix)sspiked[_idx_i];
        %(idx_type)s rk_pre = 0;

        %(idx_type)s _idx_beg = post_rank * this->num_columns_;
        %(idx_type)s _idx_end = (post_rank+1) * this->num_columns_;

        for (%(size_type)s j = _idx_beg; j < _idx_end; j++, rk_pre++) {
            if(mask_[j]) {
%(event_driven)s
%(post_event)s
            }
        }
    }
}
"""

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,
    'event_driven': event_driven,

    #operations
    'rate_coded_sum': dense_summation_operation,
    'vectorized_default_psp': {
        'sse': {
            'multi_w': continuous_transmission_sse
        },
        'avx': {
            'multi_w': continuous_transmission_avx
        },
        'avx512': {
            'multi_w': continuous_transmission_avx512
        }
    },
    'spiking_sum_fixed_delay': {
        'inner_loop': spiking_summation_fixed_delay_inner_loop,
        'outer_loop': spiking_summation_fixed_delay_outer_loop,
    },
    'spiking_sum_fixed_delay_only_psp': {
        'inner_loop': spiking_summation_fixed_delay_only_psp_inner_loop,
        'outer_loop': None
    },
    'update_variables': dense_update_variables,
    'post_event': spiking_post_event
}

conn_ids = {
    'local_index': '[j]',
    'semiglobal_index': '[i]',
    'global_index': '',
    'post_index': '[rk_post]',
    'pre_index': '[rk_pre]',
    'delay_u' : '[delay-1]' # uniform delay
}



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.ConfigManagement import ConfigManager

from .ProfileGenerator import ProfileGenerator
from .ProfileTemplate import profile_base_template, cuda_profile_template, cuda_profile_header

class CUDAProfile(ProfileGenerator):

    def __init__(self, annarchy_dir, net_id):
        ProfileGenerator.__init__(self, annarchy_dir, net_id)

    def generate(self):
        """
        Generate Profiling class code, called from Generator instance.
        """
        # Generate header for profiling
        with open(self.annarchy_dir+'/generate/net'+str(self._net_id)+'/Profiling.hpp', 'w') as ofile:
            ofile.write(self._generate_header())

    def generate_body_dict(self):
        """
        Creates a dictionary, contain profile code snippets.
        """
        body_dict = {
            'prof_include': cuda_profile_template['include'],
            'prof_step_pre': cuda_profile_template['step_pre'],
            'prof_step_post': cuda_profile_template['step_post'],
            'prof_run_pre': cuda_profile_template['run_pre'],
            'prof_run_post': cuda_profile_template['run_post'],
            'prof_proj_psp_pre': cuda_profile_template['proj_psp_pre'],
            'prof_proj_psp_post': cuda_profile_template['proj_psp_post'],
            'prof_proj_step_pre': cuda_profile_template['proj_step_pre'],
            'prof_proj_step_post': cuda_profile_template['proj_step_post'],
            'prof_neur_step_pre': cuda_profile_template['neur_step_pre'],
            'prof_neur_step_post': cuda_profile_template['neur_step_post'],
            'prof_record_pre': cuda_profile_template['record_pre'],
            'prof_record_post': cuda_profile_template['record_post']
        }
        return body_dict

    def generate_init_network(self):
        return cuda_profile_template['init']

    def generate_init_population(self, pop):
        declare = """
    Measurement* measure_step;
    Measurement* measure_gather;
"""
        init = """        // Profiling
        measure_step = prof_ptr->register_function("pop", "%(name)s", %(id)s, "step", "%(label)s");
        measure_gather = prof_ptr->register_function("pop", "%(name)s",  %(id)s, "spike", "%(label)s");
""" % {'name': pop.name, 'id': pop.id, 'label': pop.name}

        return declare, init

    def generate_init_projection(self, proj):
        """
        Generate initialization code for projection
        """
        declare = """
    Measurement* measure_psp;
    Measurement* measure_step;
    Measurement* measure_pe;
"""
        if isinstance(proj.target, str):
            target = proj.target
        else:
            target = proj.target[0]
            for tar in proj.target[1:]:
                target += "_"+tar

        init = """        // Profiling
        measure_psp = prof_ptr->register_function("proj", "%(name)s", %(id_proj)s, "psp", "%(label)s");
        measure_step = prof_ptr->register_function("proj", "%(name)s", %(id_proj)s, "step", "%(label)s");
        measure_pe = prof_ptr->register_function("proj", "%(name)s", %(id_proj)s, "post_event", "%(label)s");
""" % {'id_proj': proj.id, 'name': proj.name, 'label': proj.pre.name+'_'+proj.post.name+'_'+target}

        return declare, init

    def annotate_computesum_rate(self, proj, code):
        """
        annotate the computesum compuation code
        """
        prof_begin = cuda_profile_template['compute_psp']['before'] % {'id':proj.id, 'name': 'proj'+str(proj.id)}
        prof_end = cuda_profile_template['compute_psp']['after'] % {'id':proj.id, 'name': 'proj'+str(proj.id)}

        prof_code = """
        // first run, measuring average time
        %(prof_begin)s
%(code)s
        %(prof_end)s
""" % {'code': code,
       'prof_begin': prof_begin,
       'prof_end': prof_end
       }
        return prof_code

    def annotate_computesum_spiking(self, proj, code):
        """
        annotate the computesum compuation code
        """
        prof_begin = cuda_profile_template['compute_psp']['before'] % {'id':proj.id, 'name': 'proj'+str(proj.id)}
        prof_end = cuda_profile_template['compute_psp']['after'] % {'id':proj.id, 'name': 'proj'+str(proj.id)}

        prof_code = """
        // first run, measuring average time
        %(prof_begin)s
%(code)s
        %(prof_end)s
""" % {'code': code,
       'prof_begin': prof_begin,
       'prof_end': prof_end
       }
        return prof_code

    def annotate_update_synapse(self, proj, code):
        """
        annotate the update synapse code, generated by ProjectionGenerator.update_synapse()
        """
        prof_begin = cuda_profile_template['update_synapse']['before'] % {'id':proj.id, 'name': 'proj'+str(proj.id)}
        prof_end = cuda_profile_template['update_synapse']['after'] % {'id':proj.id, 'name': 'proj'+str(proj.id)}

        prof_code = """
// first run, measuring average time
%(prof_begin)s
%(code)s
%(prof_end)s
""" % {'code': code,
       'prof_begin': prof_begin,
       'prof_end': prof_end
       }

        return prof_code

    def annotate_post_event(self, proj, code):
        """
        annotate the post-event code
        """
        prof_begin = cuda_profile_template['post_event']['before'] % {'id':proj.id, 'name': 'proj'+str(proj.id)}
        prof_end = cuda_profile_template['post_event']['after'] % {'id':proj.id, 'name': 'proj'+str(proj.id)}

        prof_code = """
// first run, measuring average time
%(prof_begin)s
%(code)s
%(prof_end)s
""" % {'code': code,
       'prof_begin': prof_begin,
       'prof_end': prof_end
       }

        return prof_code

    def annotate_update_neuron(self, pop, code):
        """
        annotate the update neuron code
        """
        prof_begin = cuda_profile_template['update_neuron']['before'] % {'id': pop.id, 'name': pop.name}
        prof_end = cuda_profile_template['update_neuron']['after'] % {'id': pop.id, 'name': pop.name}

        prof_code = """
        // first run, measuring average time
    %(prof_begin)s
%(code)s
    %(prof_end)s
""" % {'code': code,
       'prof_begin': prof_begin,
       'prof_end': prof_end
       }
        return prof_code

    def annotate_spike_gather(self, pop, code):
        """
        annotate the update neuron code
        """
        prof_begin = cuda_profile_template['spike_gather']['before'] % {'id': pop.id, 'name': pop.name}
        prof_end = cuda_profile_template['spike_gather']['after'] % {'id': pop.id, 'name': pop.name}

        prof_code = """
        // first run, measuring average time
    %(prof_begin)s
%(code)s
    %(prof_end)s
""" % {'code': code,
       'prof_begin': prof_begin,
       'prof_end': prof_end
       }
        return prof_code

    def _generate_header(self):
        """
        generate Profiling.hpp
        """
        config_xml = """
        _out_file << "  <config>" << std::endl;
        _out_file << "    <paradigm>%(paradigm)s</paradigm>" << std::endl;
        _out_file << "  </config>" << std::endl;
        """ % {'paradigm': ConfigManager().get('paradigm', self._net_id)}

        timer_import = "#include <cuda_runtime_api.h>"
        timer_start = "cudaEvent_t _profiler_start;"
        timer_init = """
        cudaEventCreate(&_profiler_start);
        cudaEventRecord(_profiler_start);
"""

        config = ConfigManager().get('paradigm', self._net_id)
        return profile_base_template % {
            'timer_import': timer_import,
            'timer_start_decl': timer_start,
            'timer_init': timer_init,
            'config': config,
            'config_xml': config_xml,
            'measurement_class': cuda_profile_header
        }



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

class ProfileGenerator(object):
    """
    Base class to extent the generated code by profiling annotations.
    """
    def __init__(self, annarchy_dir, net_id):
        """
        Initialize ProfileGenerator.
        """
        self.annarchy_dir = annarchy_dir
        self._net_id = net_id

    def generate(self):
        """
        Called from Codegenerator, this method is responsible for
        creation of header file for the Profiling class.
        """
        raise NotImplementedError

    def generate_body_dict(self):
        """
        Creates a dictionary, contain profile code snippets. Should
        be overwritten by the specific measurement classes. This
        dict is also used in absence of profiling.
        """
        body_dict = {
            'prof_include': "",
            'prof_step_pre': "",
            'prof_step_post': "",
            'prof_run_pre': "",
            'prof_run_post': "",
            'prof_proj_psp_pre': "",
            'prof_proj_psp_post': "",
            'prof_proj_step_pre': "",
            'prof_proj_step_post': "",
            'prof_proj_post_event_pre': "",
            'prof_proj_post_event_post': "",
            'prof_neur_step_pre': "",
            'prof_neur_step_post': "",
            'prof_record_pre': "",
            'prof_record_post': "",
            'prof_rng_pre': "",
            'prof_rng_post': "",
            'prof_global_ops_pre': "",
            'prof_global_ops_post': ""
        }
        return body_dict

    def generate_include(self):
        "Implemented by child class"
        raise NotImplementedError

    def generate_init_network(self):
        "Implemented by child class"
        raise NotImplementedError

    def generate_init_population(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def generate_init_projection(self, proj):
        "Implemented by child class"
        raise NotImplementedError

    def annotate_computesum_rate(self, proj, code):
        "Implemented by child class"
        raise NotImplementedError

    def annotate_computesum_spiking(self, proj, code):
        "Implemented by child class"
        raise NotImplementedError

    def annotate_update_synapse(self, proj, code):
        "Implemented by child class"
        raise NotImplementedError

    def annotate_post_event(self, proj, code):
        "Implemented by child class"
        raise NotImplementedError

    def annotate_update_neuron(self, pop, code):
        "Implemented by child class"
        raise NotImplementedError

    def annotate_spike_cond(self, pop, code):
        "Implemented by child class"
        raise NotImplementedError

    def annotate_update_rng(self, pop, code):
        "Implemented by child class"
        raise NotImplementedError

    def annotate_update_delay(self, pop, code):
        "Implemented by child class"
        raise NotImplementedError

    def _generate_header(self):
        "Implemented by child class"
        raise NotImplementedError


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

profiler_template = """
    // Simulation kernel profiling
    nanobind::class_<Profiling>(m, "Profiling_wrapper")
        // Constructor
        .def(nanobind::init<>())
        .def("get_avg_time", &Profiling::get_avg_time)
        .def("get_std_time", &Profiling::get_std_time);
"""


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern import Messages
from ANNarchy.parser.Equation import Equation
from ANNarchy.parser.Function import FunctionParser
from ANNarchy.parser.StringManipulation import *


def translate_ITE(name, eq, condition, description, untouched, function=False, net_id=0):
    " Recursively processes the different parts of an ITE statement"

    if function:
        solver = FunctionParser
    else:
        solver = Equation

    def process_condition(condition):
        if_statement = condition[0]
        then_statement = condition[1]
        else_statement = condition[2]

        if_solver = solver(name, if_statement, description,
                          untouched = untouched.keys(),
                          type='cond',
                          net_id=net_id)
        if_code = if_solver.parse()
        if_deps = if_solver.dependencies()

        if isinstance(then_statement, list): # nested conditional
            then_code, then_deps =  process_condition(then_statement)
        else:
            then_solver = solver(name, then_statement, description,
                          untouched = untouched.keys(),
                          type='return',
                          net_id=net_id)
            then_code = then_solver.parse().split(';')[0]
            then_deps = then_solver.dependencies()
        
        if isinstance(else_statement, list): # nested conditional
            else_code, else_deps =  process_condition(else_statement)
        else:
            else_solver = solver(name, else_statement, description,
                          untouched = untouched.keys(),
                          type='return',
                          net_id=net_id)
            else_code = else_solver.parse().split(';')[0]
            else_deps = else_solver.dependencies()

        code = '(' + if_code + ' ? ' + then_code + ' : ' + else_code + ')'
        deps = list(set(if_deps + then_deps + else_deps))
        return code, deps

    # Main equation, where the right part is __conditional__
    translator = solver(name, eq, description,
                          untouched = untouched.keys(),
                          net_id=net_id)
    code = translator.parse()
    deps = translator.dependencies()

    # Process the (possibly multiple) ITE
    for i in range(len(condition)):
        itecode, itedeps =  process_condition(condition[i])
        deps += itedeps

        # Replace
        if isinstance(code, str):
            code = code.replace('__conditional__'+str(i), itecode)
        else:
            code[1] = code[1].replace('__conditional__'+str(i), itecode)

    deps = list(set(deps)) # remove doublings
    return code, deps


def extract_ite(name, eq, description, split=True):
    """ Extracts if-then-else statements and processes them.

    If-then-else statements must be of the form:

    .. code-block:: python

        variable = if condition: ...
                       val1 ...
                   else: ...
                       val2

    Conditional statements can be nested, but they should return only one value!
    """

    def transform(code):
        " Transforms the code into a list of lines."
        res = []
        items = []
        for arg in code.split(':'):
            items.append( arg.strip())
        for i in range(len(items)):
            if items[i].startswith('if '):
                res.append( items[i].strip() )
            elif items[i].endswith('else'):
                res.append(items[i].split('else')[0].strip() )
                res.append('else' )
            else: # the last then
                res.append( items[i].strip() )
        return res


    def parse(lines):
        " Recursive analysis of if-else statements"
        result = []
        while lines:
            if lines[0].startswith('if'):
                block = [lines.pop(0).split('if')[1], parse(lines)]
                if lines[0].startswith('else'):
                    lines.pop(0)
                    block.append(parse(lines))
                result.append(block)
            elif not lines[0].startswith(('else')):
                result.append(lines.pop(0))
            else:
                break
        return result[0]

    # If no if, not a conditional
    if not 'if ' in eq:
        return eq, []

    # Process the equation
    condition = []
    # Eventually split around =
    if split:
        left, right =  eq.split('=', 1)
    else:
        left = ''
        right = eq

    nb_then = len(re.findall(r':', right))
    nb_else = len(re.findall(r'else', right))
    # The equation contains a conditional statement
    if nb_then > 0:
        # A if must be right after the equal sign
        if not right.strip().startswith('if'):
            Messages._error(eq, '\nThe right term must directly start with a if statement.')

        # It must have the same number of : and of else
        if not nb_then == 2*nb_else:
            Messages._error(eq, '\nConditional statements must use both : and else.')

        multilined = transform(right)
        condition = parse(multilined)
        right = ' __conditional__0 ' # only one conditional allowed in that case
        if split:
            eq = left + '=' + right
        else:
            eq = right
    else:
        Messages._print(eq)
        Messages._error('Conditional statements must define "then" and "else" values.\n var = if condition: a else: b')

    return eq, [condition]



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import re

from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern import Messages
from ANNarchy.models.Synapses import DefaultRateCodedSynapse

# No variable can have these names
reserved_variables = [
    't',
    'dt',
    't_pre',
    't_post',
    't_last',
    'last_spike',
    'rk_post',
    'rk_pre',
    'i',
    'j',
    'active',
    'refractory',
    'size',
]

def check_structure(populations, projections):
    """
    Checks the structure before compilation to display more useful error messages.
    """
    from ANNarchy.extensions.convolution.Transpose import Transpose

    # Check variable names
    _check_reserved_names(populations, projections)

    # Check that projections are created before compile
    for proj in projections:
        
        if isinstance(proj, Transpose):
            continue

        if not proj._connection_method:
            Messages._error('The projection between populations', proj.pre.id, 'and', proj.post.id, 'has not been connected.',
                            ' Call a connector method before compiling the network.')

    # Check if the storage formats are valid for the selected paradigm
    _check_storage_formats(projections)

    # Check that synapses access existing variables in the pre or post neurons
    _check_prepost(populations, projections)

    # Check locality of variable is respected
    _check_locality(populations, projections)

    # Check structural plasticity
    _check_structural_plasticity(projections)

def check_experimental_features(populations, projections):
    """
    The idea behind this method, is to check if new experimental features are used. This
    should help also the user to be aware of changes.
    """
    net_id = populations[0].net_id

    detected_formats = []
    for proj in projections:
        detected_formats.append((proj._storage_format, proj._storage_order, proj.synapse_type.type))
    detected_formats = list(set(detected_formats))

    # CPU-related formats
    if ConfigManager().get('paradigm', net_id) == "openmp":
        if ConfigManager().get("disable_SIMD_SpMV", net_id) == False:
            Messages._warning("Using hand-written SIMD kernel for continuous transmission is an experimental feature, we greatly appreciate bug reports.")

        for fmt in detected_formats:
            if fmt == "lil":
                # nothing to tell, its the default
                continue

            elif fmt[0] == "csr" and fmt[1] == "pre_to_post":
                Messages._warning("Compressed sparse row (CSR) and pre_to_post ordering representation is an experimental feature, we greatly appreciate bug reports.")

            elif fmt[0] == "bsr":
                Messages._warning("Blocked sparse row (BSR) representation is an experimental feature, we greatly appreciate bug reports.")

            elif fmt[0] == "coo":
                Messages._warning("Coordinate (COO) representation is an experimental feature, we greatly appreciate bug reports.")

            elif fmt[0] == "dia":
                Messages._warning("Diagonal (dia) representation is an experimental feature, we greatly appreciate bug reports.")

            elif fmt[0] == "ellr":
                Messages._warning("ELLPACK-R (ELLR) representation is an experimental feature, we greatly appreciate bug reports.")

            elif fmt[0] == "sell":
                Messages._warning("Sliced ELLPACK (SELL) representation is an experimental feature, we greatly appreciate bug reports.")

            elif fmt[0] == "ell":
                Messages._warning("ELLPACK (ELL) representation is an experimental feature, we greatly appreciate bug reports.")

            elif fmt[0] == "hyb":
                Messages._warning("Hybrid (ELL + COO) representation is an experimental feature, we greatly appreciate bug reports.")

            elif fmt[0] == "dense" and fmt[2]=="spike":
                Messages._warning("Dense representation is an experimental feature for spiking models, we greatly appreciate bug reports.")

            else:   # check invalid arguments as last
                if fmt[0] not in ["lil", "csr", "bsr", "coo", "dia", "ellr", "sell", "ell", "hyb", "dense"]:
                    Messages._error("Invalid storage format provided for execution on CPUs:", fmt[0])
                if fmt[1] not in ["post_to_pre", "pre_to_post"]:
                    Messages._error("Invalid storage order provided:", fmt[1])

    # GPU-related formats
    elif ConfigManager().get('paradigm', net_id) == "cuda":

        for fmt in detected_formats:
            if fmt == "lil" or fmt == "csr":
                # nothing to tell, its the default
                continue

            elif fmt[0] == "dense":
                Messages._warning("Dense representation is an experimental feature, we greatly appreciate bug reports.")

            elif proj._storage_format == "sell":
                Messages._warning("Sliced ELLPACK representation is an experimental feature, we greatly appreciate bug reports.")

            elif proj._storage_format == "ellr":
                Messages._warning("ELLPACK-R (ELLR) representation is an experimental feature, we greatly appreciate bug reports.")

            elif proj._storage_format == "bsr":
                Messages._warning("Blocked sparse row (BSR) representation is an experimental feature, we greatly appreciate bug reports.")

            elif proj._storage_format == "coo":
                Messages._warning("Coordinate (COO) representation is an experimental feature, we greatly appreciate bug reports.")

            elif proj._storage_format == "hyb":
                Messages._warning("Hybrid (ELL + COO) representation is an experimental feature, we greatly appreciate bug reports.")

            else:   # check invalid arguments as last
                if fmt[0] not in ["lil", "csr", "bsr", "coo", "ellr", "sell", "hyb", "dense"]:
                    Messages._error("Invalid storage format provided for execution on CPUs GPUs:", fmt[0])
                if fmt[1] not in ["post_to_pre", "pre_to_post"]:
                    Messages._error("Invalid storage order provided:", fmt[1])

    else:
        pass

def _check_reserved_names(populations, projections):
    """
    Checks no reserved variable names is redefined
    """
    # Check populations
    for pop in populations:
        # Reserved variable names
        for term in reserved_variables:
            if term in pop.attributes:
                Messages._print(pop.neuron_type.parameters)
                Messages._print(pop.neuron_type.equations)
                Messages._error(term + ' is a reserved variable name')

    # Check projections
    for proj in projections:
        # Reserved variable names
        for term in reserved_variables:
            if term in proj.attributes:
                Messages._print(proj.synapse_type.parameters)
                Messages._print(proj.synapse_type.equations)
                Messages._error(term + ' is a reserved variable name')

def _check_storage_formats(projections):
    """
    ANNarchy 4.7 introduced a set of sparse matrix formats. Some of them are not implemented for
    all paradigms or might not support specific optimizations.
    """
    for proj in projections:
        # Most of the sparse matrix formats are not trivially invertable and therefore we can not implement
        # spiking models with them
        if proj.synapse_type.type == "spike" and proj._storage_format in ["csr_vector", "csr_scalar", "ell", "ellr", "coo", "hyb", "dia"]:
            raise Messages.InvalidConfiguration("Using 'storage_format="+ proj._storage_format + "' is not allowed for spiking synapses.")

        # Continous signal transmission in spiking models, e.g. gap junctions, should not be combined with pre-to-post
        if proj.synapse_type.type == "spike":
            if 'psp' in  proj.synapse_type.description.keys() and proj._storage_order=="pre_to_post":
                raise Messages.InvalidConfiguration("Using continuous transmission within a spiking synapse prevents the application of pre-to-post matrix ordering")

        # For some of the sparse matrix formats we don't implemented plasticity for rate-coded models yet.
        if proj.synapse_type.type == "rate":
            if _check_paradigm("openmp", proj.net_id):
                if proj._storage_format in ["coo", "hyb", "bsr", "sell", "dia"] and not isinstance(proj.synapse_type, DefaultRateCodedSynapse):
                    raise Messages.InvalidConfiguration("Using 'storage_format="+ proj._storage_format + "' is only allowed for default rate-coded synapses yet.")
            elif _check_paradigm("cuda", proj.net_id):
                if proj._storage_format in ["coo", "hyb", "bsr", "ell", "sell", "csr_vector", "csr_scalar"] and not isinstance(proj.synapse_type, DefaultRateCodedSynapse):
                    raise Messages.InvalidConfiguration("Using 'storage_format="+ proj._storage_format + "' is only allowed for default rate-coded synapses yet.")

        # Single weight optimization available?
        if proj._has_single_weight() and proj._storage_format in ["dense", "bsr"]:
            raise Messages.InvalidConfiguration("Using 'storage_format="+ proj._storage_format + "' is not allowed for single weight projections.")

        # In some cases we don't allow the usage of non-unifom delay
        if (proj.max_delay > 1 and proj.uniform_delay == -1):
            if _check_paradigm("cuda", proj.net_id):
                raise Messages.InvalidConfiguration("Using non-uniform delays is not available for CUDA devices.")

            else:
                if proj._storage_format == "ellr":
                    raise Messages.InvalidConfiguration("Using 'storage_format="+ proj._storage_format + "' is and non-uniform delays is not implemented.")

        if proj._storage_format == "dia":
            if _check_paradigm("cuda", proj.net_id):
                raise Messages.InvalidConfiguration('Using diagonal format is limited to CPUs yet.')

            if proj.pre.size < proj.post.size:
                raise Messages.InvalidConfiguration('Using diagonal format is not implemented for projections where the pre-synaptic layer is smaller than the post-synaptic one.')

            if isinstance(proj.post, PopulationView):
                raise Messages.InvalidConfiguration('Using diagonal format and post-synaptic PopulationViews is not available.')

        if not _check_paradigm("cuda", proj.net_id) and (proj._storage_format in ["csr_scalar", "csr_vector"]):
            Messages._error("The CSR variants csr_scalar/csr_vector are only intended for GPUs.")

        if _check_paradigm("cuda", proj.net_id) and proj._storage_format == "lil":
            proj._storage_format = "csr"
            if not isinstance(proj, SpecificProjection):
                Messages._info("LIL-type projections are not available for GPU devices ... default to CSR")

        if _check_paradigm("cuda", proj.net_id) and proj._storage_format == "ell":
            Messages._info("We would recommend to use ELLPACK-R (format=ellr) on GPUs.")
        
def _check_prepost(populations, projections):
    """
    Checks that when a synapse uses pre.x r post.x, the variable x exists in the corresponding neuron
    """
    for proj in projections:

        for dep in  proj.synapse_type.description['dependencies']['pre']:
            if dep.startswith('sum('):
                target = re.findall(r'\(([\s\w]+)\)', dep)[0].strip()
                if not target in proj.pre.targets:
                    Messages._print(proj.synapse_type.equations)
                    Messages._error('The pre-synaptic population ' + proj.pre.name + ' receives no projection with the type ' + target)
                continue

            if not dep in proj.pre.attributes:
                Messages._print(proj.synapse_type.equations)
                Messages._error('The pre-synaptic population ' + proj.pre.name + ' has no variable called ' + dep)

        for dep in proj.synapse_type.description['dependencies']['post']:
            if dep.startswith('sum('):
                target = re.findall(r'\(([\s\w]+)\)', dep)[0].strip()
                if not target in proj.post.targets:
                    Messages._print(proj.synapse_type.equations)
                    Messages._error('The post-synaptic population ' + proj.post.name + ' receives no projection with the type ' + target)
                continue

            if not dep in proj.post.attributes:
                Messages._print(proj.synapse_type.equations)
                Messages._error('The post-synaptic population ' + proj.post.name + ' has no variable called ' + dep)


def _check_locality(populations, projections):
    """
    Checks that a global variable does not depend on local ones.
    """
    for proj in projections:

        for var in proj.synapse_type.description['variables']:

            if var['locality'] == 'global': # cannot depend on local or semiglobal variables
                # Inside the equation
                for v in var['dependencies']:
                    if _get_locality(v, proj.synapse_type.description) in ['local', 'semiglobal']:
                        Messages._print(var['eq'])
                        Messages._error('The global variable', var['name'], 'cannot depend on a synapse-specific/post-synaptic one:', v)

                # As pre/post dependencies
                deps = var['prepost_dependencies']
                if len(deps['pre']) > 0 or len(deps['post']) > 0 : 
                    Messages._print(proj.synapse_type.equations)
                    Messages._error('The global variable', var['name'], 'cannot depend on pre- or post-synaptic variables.')

            if var['locality'] == 'semiglobal': # cannot depend on pre-synaptic variables
                # Inside the equation
                for v in var['dependencies']:
                    if _get_locality(v, proj.synapse_type.description) == 'local':
                        Messages._print(var['eq'])
                        Messages._error('The postsynaptic variable', var['name'], 'cannot depend on a synapse-specific one:', v)

                # As pre/post dependencies
                deps = var['prepost_dependencies']
                if len(deps['pre']) > 0  : 
                    Messages._print(proj.synapse_type.equations)
                    Messages._error('The postsynaptic variable', var['name'], 'cannot depend on pre-synaptic ones (e.g. pre.r).')


def _get_locality(name, description):
    "Returns the locality of an attribute based on its name"
    for var in description['variables'] + description['parameters']:
        if var['name'] == name:
            return var['locality']
    return 'local'

def _check_structural_plasticity(projections:list["Projection"]):
    "If a synapse implements structural plasticity, set the config flag to True."

    for proj in projections:
        if proj.synapse_type.creating or proj.synapse_type.pruning:
            if _check_paradigm("cuda", proj.net_id):
                # HD: this could be relaxed in future and only limited to using equation-based structural plasticity.
                #     As Dendrite.create_-/prune_synapse calls could be applied on host-side and then update device before simulate().
                raise Messages.InvalidConfiguration("Structural plasticity is not supported on GPU devices.")

            ConfigManager().set('structural_plasticity', True, proj.net_id)